import logging
from typing import Any

import regex as re
from contextlib import suppress
from collections.abc import Iterator
from urllib.parse import unquote

TEMPLATE_NAMES_PATTERN = re.compile("\\{([\\w\\?;][-\\w\\.,]*)\\}")


def tvars(route: str, url: str) -> dict[str, Any]:
    """
    Get the variables from a URL given the route template. If the URL does not match the given route, then return {}.

    :param route: the route template.
    :param url: the url that may be a valid one for the given route.
    :return: a dictionary containing references to all the variables from the route that were given in curly braces.
    The variable values are decoded if necessary.
    :raises ValueError: if an error occurs trying to compute the variables.
    """
    try:
        pattern = _to_regex(route)
        match = pattern.fullmatch(url, partial=True)
        while match and match.partial:
            route = route[:-1]
            with suppress(Exception):
                match = _to_regex(route).fullmatch(url, partial=True)
        return {k: unquote(v) for k, v in match.groupdict().items()} if match else {}
    except re.error as e:
        raise ValueError from e

def _to_regex(route: str) -> re.Pattern:
    """
    Convert a route to a regex pattern.

    For each variable in the route, a capture group is added. Everything else is an exact match.

    :param route: the route template.
    :return: the generated regex pattern.
    """
    metadata_ = _metadata(route)
    l = []
    for m in metadata_:
        if m[3]:
            l.append(r'(?P<' + m[0] + r'>.+?)')
        else:
            l.append(m[0])
    return re.compile(''.join(l) + '(?:$|/.*)')


def _metadata(route: str) -> Iterator[tuple[str, int, int, bool]]:
    """
    Splits the route into components such that the variables and constants are separate. The function returns an
    iterator of components.

    The first index of the yielded content is the content itself. The second and third represents the start and end
    of a range that indicates the indices of the content. The fourth index is a boolean that is True if the content
    is a variable.

    :param route: the route template.
    :return: an iterator of 4-tuples containing the content, the start index, the end index (exclusive), and whether
    it is a variable.
    """
    if route is None:
        raise ValueError('route cannot be None')
    m = {match.group(): (match.start(), match.end()) for match in
         TEMPLATE_NAMES_PATTERN.finditer(route)}
    split_ = TEMPLATE_NAMES_PATTERN.split(route)
    start = 0
    i = 0
    for k, v in m.items():
        if v[0] > start:
            yield split_[i], start, v[0], False
            start = v[1]
            i += 1
        yield split_[i], v[0], v[1], True
        i += 1
    else:
        route_len = len(route)
        if route_len > start:
            yield split_[i], start, route_len, False



import heaobject.root

from heaserver.service.util import queued_processing
from .. import response
from ..heaobjectsupport import new_heaobject_from_type, RESTPermissionGroup
from ..appproperty import HEA_DB, HEA_CACHE
from ..aiohttp import extract_sort_attrs, extract_sorts, SortOrder
from .mongo import MongoContext, Mongo
from heaobject.error import DeserializeException
from aiohttp.web import Request, StreamResponse, Response, HTTPError
from typing import Any, AsyncGenerator, Literal, IO, overload, Optional
from heaobject.root import DesktopObject, DesktopObjectDict, desktop_object_from_dict, Permission, DesktopObjectTypeVar
from heaobject.user import NONE_USER
from heaserver.service.oidcclaimhdrs import SUB
from pymongo.errors import WriteError, DuplicateKeyError
from collections.abc import Sequence, Mapping, Callable, Awaitable, Iterator
from asyncio import gather
from enum import Enum
from copy import deepcopy
from itertools import chain, repeat
import logging


SortDict = dict[str, Literal[-1, 1]]
SortDictCacheKey = tuple[tuple[str, Literal[-1, 1]], ...]


async def get_dict(request: Request, collection: str, volume_id: str | None = None, mongo: Mongo | None = None,
                   context: heaobject.root.PermissionContext | None = None) -> DesktopObjectDict | None:
    """
    Gets the requested desktop object as a desktop object dict.

    :param request: the HTTP request, which must have an id path parameter. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :param mongo: the Mongo context. If None, a new Mongo context is created.
    :param context: the permission context. If None, a new context is created.
    :return: a desktop object dict or None if the object was not found.
    """
    sub = request.headers.get(SUB, NONE_USER)
    cache_key = (sub, collection, f'id^{request.match_info["id"]}')
    cached_value = request.app[HEA_CACHE].get(cache_key)
    if cached_value is not None:
        return deepcopy(cached_value)
    else:
        async def fetch_from_mongo(mongo: Mongo) -> DesktopObjectDict | None:
            context_ = context if context else mongo.get_default_permission_context(request)
            result = await mongo.get(request, collection, var_parts='id', context=context_)

            if result is not None:
                obj = heaobject.root.desktop_object_from_dict(result)
                permitted = await RESTPermissionGroup.GETTER_PERMS.has_any(obj, context_)
                if not permitted:
                    return None
                request.app[HEA_CACHE][cache_key] = result
                return obj.to_dict()
            else:
                return None
        if mongo:
            return await fetch_from_mongo(mongo)
        else:
            async with MongoContext(request, volume_id) as mongo:
                return await fetch_from_mongo(mongo)


@overload
async def get_desktop_object(request: Request, collection: str, *, volume_id: str | None = None,
                             obj: DesktopObject | None = None,
                             mongo: Mongo | None = None,
                             context: heaobject.root.PermissionContext | None = None) -> DesktopObject | None:
    ...

@overload
async def get_desktop_object(request: Request, collection: str, *,
                             type_: type[DesktopObjectTypeVar], volume_id: str | None = None,
                             obj: DesktopObjectTypeVar | None = None,
                             mongo: Mongo | None = None,
                             context: heaobject.root.PermissionContext | None = None) -> DesktopObjectTypeVar | None:
    ...

async def get_desktop_object(request: Request, collection: str, *,
                             type_ = DesktopObject, volume_id: str | None = None,
                             obj: DesktopObject | None = None,
                             mongo: Mongo | None = None,
                             context: heaobject.root.PermissionContext | None = None) -> DesktopObject | None:
    """
    Gets the desktop object with the specified id.

    :param request: the HTTP request, which must have an id value in the match_info mapping. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: optional object type.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :param obj: the desktop object to populate from mongo.
    :param mongo: the Mongo context. If None, a new Mongo context is created.
    :param context: the permission context. If None, a new context is created.
    :return: a Response with the requested HEA object or Not Found.
    :raises TypeError: if the object retrieved from Mongo does not match the value of the type_ argument, or if the obj
    argument is not an instance of the type_ argument.
    """
    obj_dict = await get_dict(request, collection, volume_id, mongo=mongo, context=context)
    return _dict_to_desktop_object(obj_dict, type_=type_, obj=obj)


async def get(request: Request, collection: str, volume_id: str | None = None,
              context: heaobject.root.PermissionContext | None = None) -> Response:
    """
    Gets an HTTP response with the desktop object with the specified id in the body. The desktop object is
    formatted according to the requested mime types in the HTTP request's
    Accept header as one of the formats supported by the
    heaserver.service.representor module.

    :param request: the HTTP request, which must have an id value in the match_info mapping. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :param context: the PermissionContext. If None, a new context is created.
    :return: a Response with the requested HEA object or Not Found.
    """
    async with MongoContext(request, volume_id) as mongo:
        context = context or mongo.get_default_permission_context(request)
        result = await get_dict(request, collection, volume_id, mongo=mongo, context=context)
        if result is None:
            return await response.get(request, None)
        else:
            obj = desktop_object_from_dict(result)
            return await response.get(request, obj.to_dict(), await obj.get_permissions(context),
                                    await obj.get_all_attribute_permissions(context))


async def get_content(request: Request, collection: str, volume_id: str | None = None) -> StreamResponse:
    """
    Gets the HEA object's associated content.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: an aiohttp StreamResponse with the requested HEA object or Not Found.
    """
    sub = request.headers.get(SUB, NONE_USER)
    async with MongoContext(request, volume_id) as mongo:
        context = mongo.get_default_permission_context(request)
        result = await mongo.get(request, collection, var_parts='id', context=context)

        if result is not None:
            obj = heaobject.root.desktop_object_from_dict(result)
            permitted = await RESTPermissionGroup.GETTER_PERMS.has_any(obj, context)
            if not permitted:
                return response.status_not_found()
        out = await mongo.get_content(request, collection, var_parts='id', context=context)
        if out is not None:
            return await response.get_streaming(request, out, 'text/plain')
        else:
            return response.status_not_found()


async def get_by_name(request: Request, collection: str,
                      volume_id: str | None = None) -> Response:
    """
    Gets an HTTP response object with the requested desktop object in the body.
    The desktop object is formatted according to the requested mime types in
    the HTTP request's Accept header as one of the formats supported by the
    heaserver.service.representor module.

    :param request: the HTTP request, which must have a name value in the match_info mapping. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested
    desktop object. If None, the root volume is assumed.
    :return: a Response with the requested desktop object or Not Found.
    """
    async with MongoContext(request, volume_id) as mongo:
        result = await get_by_name_dict(request, collection, volume_id, mongo=mongo)
        if result is None:
            return await response.get(request, None)
        else:
            obj = desktop_object_from_dict(result)
            context = mongo.get_default_permission_context(request)
            return await response.get(request, desktop_object_from_dict(result).to_dict() if result is not None else None,
                                    permissions=await obj.get_permissions(context),
                                    attribute_permissions=await obj.get_all_attribute_permissions(context))

async def get_by_name_dict(request: Request, collection: str,
                           volume_id: str | None = None, mongo: Mongo | None = None,
                           context: heaobject.root.PermissionContext | None = None) -> DesktopObjectDict | None:
    """
    Gets the requested desktop object as a desktop object dict.

    :param request: the HTTP request, which must have a name value in the match_info mapping. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA
    object. If None, the root volume is assumed.
    :param mongo: the Mongo context. If None, a new Mongo context is created.
    :param context: the permission context. If None, a new context is created.
    :return: a desktop object dict or None if the object was not found.
    """
    sub = request.headers.get(SUB, NONE_USER)
    cache_key = (sub, collection, f'name^{request.match_info["name"]}')
    cached_value = request.app[HEA_CACHE].get(cache_key)
    if cached_value is not None:
        return cached_value
    else:
        async def fetch_from_mongo(mongo: Mongo) -> DesktopObjectDict | None:
            context_ = context if context else mongo.get_default_permission_context(request)
            result = await mongo.get(request, collection, var_parts='name', context=context_)

            if result is not None:
                obj = heaobject.root.desktop_object_from_dict(result)
                permitted = await RESTPermissionGroup.GETTER_PERMS.has_any(obj, context_)
            else:
                permitted = None
            if not permitted:
                return None
            request.app[HEA_CACHE][cache_key] = result
            return result

        if mongo:
            return await fetch_from_mongo(mongo)
        else:
            async with MongoContext(request, volume_id) as mongo:
                return await fetch_from_mongo(mongo)


@overload
async def get_by_name_desktop_object(request: Request, collection: str, *, volume_id: str | None = None,
                             obj: DesktopObject | None = None,
                             mongo: Mongo | None = None,
                             context: heaobject.root.PermissionContext | None = None) -> DesktopObject | None:
    ...


@overload
async def get_by_name_desktop_object(request: Request, collection: str, *,
                             type_: type[DesktopObjectTypeVar], volume_id: str | None = None,
                             obj: DesktopObjectTypeVar | None = None,
                             mongo: Mongo | None = None,
                             context: heaobject.root.PermissionContext | None = None) -> DesktopObjectTypeVar | None:
    ...


async def get_by_name_desktop_object(request: Request, collection: str, *,
                                     type_ = DesktopObject, volume_id: str | None = None,
                                     obj: DesktopObject | None = None,
                                     mongo: Mongo | None = None,
                                     context: heaobject.root.PermissionContext | None = None) -> DesktopObject | None:
    """
    Gets the desktop object with the specified name.

    :param request: the HTTP request, which must have a name value in the match_info mapping. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: optional object type.
    :param volume_id: the id of the volume containing the requested HEA object. If None, the root volume is assumed.
    :param obj: the desktop object to populate from mongo.
    :param mongo: the Mongo context. If None, a new Mongo context is created.
    :param context: the permission context. If None, a new context is created.
    :return: a Response with the requested HEA object or Not Found.
    :raises TypeError: if the object retrieved from Mongo does not match the value of the type_ argument, or if the obj
    argument is not an instance of the type_ argument.
    """
    obj_dict = await get_by_name_dict(request, collection, volume_id, mongo=mongo, context=context)
    return _dict_to_desktop_object(obj_dict, type_=type_, obj=obj)


async def get_all(request: Request,
                  collection: str,
                  volume_id: str | None = None,
                  mongoattributes: Any | None = None,
                  sort: SortDict | None = None,
                  context: heaobject.root.PermissionContext | None = None) -> Response:
    """
    Gets an HTTP response with all requested desktop objects in the body. The desktop objects are formatted according
    to the requested mime types in the HTTP request's Accept header as one of the formats supported by the
    heaserver.service.representor module.  This function caches its results. The cache key it uses is
    (sub, collection, None, tuple of sort key-value tuples, and the cache value it stores is
    (list of desktop object dicts, list of object permissions, list of attribute permissions).

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :param sort: the sort order of the desktop objects as a dict of desktop object attribute to 1 or -1 for ascending
    or descending.
    :param context: the PermissionContext. If None, a new context is created.
    :return: a Response with a list of HEA object dicts. If no desktop objects are found, the body will contain an
    empty list.
    """
    sub = request.headers.get(SUB, NONE_USER)
    if mongoattributes is None and not request.query:
        cache_key = (sub, collection, None, tuple((key, val) for key, val in (sort or {}).items()))
        cached_value = request.app[HEA_CACHE].get(cache_key)
    else:
        cached_value = None
    if mongoattributes is None and cached_value is not None:
        return await response.get_all(request, data=cached_value[0], permissions=cached_value[1],
                                      attribute_permissions=cached_value[2])
    else:
        l: list[DesktopObjectDict] = []
        async with MongoContext(request, volume_id) as mongo:
            context = context or mongo.get_default_permission_context(request)
            gen = get_all_desktop_objects_gen(request, collection,
                                              volume_id=volume_id, mongoattributes=mongoattributes, sort=sort,
                                              mongo=mongo, context=context)
            try:
                perms: list[list[Permission]] = []
                attr_perms: list[dict[str, list[Permission]]] = []
                async def the_iter():
                    async for obj in gen:
                        l.append(obj.to_dict())
                        yield obj
                buffer: list[DesktopObject] = []
                async def process_buffer(buffer: list[DesktopObject]):
                    async def process_obj(obj: DesktopObject):
                        return await obj.get_permissions(context), await obj.get_all_attribute_permissions(context)
                    results = await gather(*(process_obj(o) for o in buffer))
                    for perm, attr_perm in results:
                        perms.append(perm)
                        attr_perms.append(attr_perm)
                async def process_desktop_object(obj: DesktopObject):
                    buffer.append(obj)
                    if len(buffer) >= 50:
                        await process_buffer(buffer)
                        buffer.clear()
                await queued_processing(the_iter(), process_desktop_object, num_workers=1)
                if buffer:
                    await process_buffer(buffer)
                if mongoattributes is None and not request.query:
                    request.app[HEA_CACHE][cache_key] = (l, perms, attr_perms)
                return await response.get_all(request, l, permissions=perms, attribute_permissions=attr_perms)
            finally:
                await gen.aclose()

async def get_all_dict(request: Request,
                       collection: str,
                       volume_id: str | None = None,
                       mongoattributes: Any | None = None,
                       sort: SortDict | None = None,
                       mongo: Mongo | None = None,
                       context: heaobject.root.PermissionContext | None = None) -> list[DesktopObjectDict]:
    """
    Gets all HEA objects as a list of desktop object dicts. This function does not cache its results, but if there are
    cached results from a prior get_all function call, it will use them. The cache key it looks for is
    (sub, collection, None, tuple of sort key-value tuples)

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA
    object. If None, the root volume is assumed.
    :param sort: the sort order of the desktop objects as a dict of desktop
    object attribute to 1 or -1 for ascending or descending.
    :param mongo: the Mongo context. If None, a new Mongo context is created.
    :param context: the PermissionContext. If None, a new context is created.
    :return: a list of DesktopObjectDict. If no desktop objects are found, the
    return value will be an empty list.
    """
    sub = request.headers.get(SUB, NONE_USER)
    if mongoattributes is None and not request.query:
        cache_key = (sub, collection, None, tuple((key, val) for key, val in (sort or {}).items()))
        cached_value = request.app[HEA_CACHE].get(cache_key)
    else:
        cached_value = None
    if mongoattributes is None and cached_value is not None:
        return deepcopy(cached_value[0])
    else:
        return [obj.to_dict() async for obj in get_all_desktop_objects_gen(request, collection, volume_id=volume_id,
                                                                        mongoattributes=mongoattributes, sort=sort,
                                                                        mongo=mongo, context=context)]


async def opener(request: Request, collection: str, volume_id: str | None = None,
                 include_desktop_object: bool = True) -> Response:
    """
    Gets choices for opening an HEA desktop object's content.

    :param request: the HTTP request. Required. If an Accepts header is provided, MIME types that do not support links
    will be ignored.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: a Response object with status code 300, and a body containing the HEA desktop object and links
    representing possible choices for opening the HEA desktop object; or Not Found.
    """
    sub = request.headers.get(SUB, NONE_USER)
    cache_key = (sub, collection, f"id^{request.match_info['id']}")
    cached_value = request.app[HEA_CACHE].get(cache_key)
    if include_desktop_object is None:
        include_desktop_object_ = True
    else:
        include_desktop_object_ = bool(include_desktop_object)
    if cached_value is not None:
        return await response.get_multiple_choices(request, cached_value if include_desktop_object_ else None)
    else:
        async with MongoContext(request, volume_id) as mongo:
            context = mongo.get_default_permission_context(request)
            result = await mongo.get(request, collection, var_parts='id', context=context)
            if result is not None:
                obj = heaobject.root.desktop_object_from_dict(result)
                if not await RESTPermissionGroup.GETTER_PERMS.has_any(obj, context):
                    return response.status_not_found()
                request.app[HEA_CACHE][cache_key] = result
                return await response.get_multiple_choices(request, result if include_desktop_object_ else None)
            else:
                return response.status_not_found()



async def post(request: Request, collection: str, type_: type[DesktopObject], default_content: IO | None = None,
               volume_id: str | None = None, resource_base: str | None = None) -> Response:
    """
    Posts the desktop object from a request body.

    :param request: the HTTP request containing the desktop object. The object's owner and the request's OIDC_CLAIM_sub
    header must match.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required. The type of the object in the request body is compared to this type
    for whether the object is an instance of the given type, and a response with a 400 status code is returned if it is
    not.
    :param default_content: an optional blank document or other default content as a file-like object. This must be not-None
    for any microservices that manage content.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :param resource_base: the base path of the created resource. If None, the collection is used as the base path.
    :return: a Response object with a status of Created and the object's URI in the Location header, otherwise returns
    a Response with an appropriate status code.
    """
    logger = logging.getLogger(__name__)
    sub = request.headers.get(SUB, NONE_USER)
    try:
        obj = await new_heaobject_from_type(request, type_)
    except DeserializeException as e:
        return response.status_bad_request(str(e))
    if not await request.app[HEA_DB].is_creator(request, type(obj)):
        logger.debug('Permission denied to %s creating object %s', sub, obj)
        return response.status_forbidden(f'Permission denied creating object of type {type_.get_type_name()}')
    if obj.owner != sub:
        logger.debug('Permission denied to %s creating object with owner %s', sub, obj.owner)
        return response.status_forbidden(f'Permission denied creating object of type {type_.get_type_name()}')
    async with MongoContext(request, volume_id) as mongo:
        try:
            result = await mongo.post(request, obj, collection, default_content)
            to_delete = []
            for cache_key in request.app[HEA_CACHE]:
                if cache_key[1] == collection and cache_key[2] is None:
                    to_delete.append(cache_key)
            for cache_key in to_delete:
                request.app[HEA_CACHE].pop(cache_key, None)
            return await response.post(request, result, resource_base if resource_base is not None else collection)
        except DuplicateKeyError as e:
            return response.status_conflict(f'Object {obj.display_name} already exists')

async def post_dict_return_id(request: Request, obj_dict: DesktopObjectDict, collection: str,
                              type_: type[DesktopObject], default_content: IO | None = None,
                              volume_id: str | None = None) -> str | None:
    """
    Posts the desktop object from a request body.

    :param request: the HTTP request containing the desktop object. The object's owner and the request's OIDC_CLAIM_sub
    header must match.
    :param obj_dict: the object to create (required).
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required. The type of the object in the request body is compared to this type
    for whether the object is an instance of the given type, and a response with a 400 status code is returned if it is
    not.
    :param default_content: an optional blank document or other default content as a file-like object. This must be not-None
    for any microservices that manage content.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: the created object's id.
    :raises HTTPError: if an error occurs creating the object on the server.
    """
    logger = logging.getLogger(__name__)
    sub = request.headers.get(SUB, NONE_USER)
    owner = obj_dict.get('owner')
    if sub != owner:
        logger.debug('Permission denied to %s creating object %s', sub, obj_dict)
        raise response.status_forbidden(f'Permission denied creating object with owner {owner}')
    try:
        obj = desktop_object_from_dict(obj_dict)
    except (DeserializeException, ValueError, TypeError) as e:
        raise response.status_bad_request(str(e))
    async with MongoContext(request, volume_id) as mongo:
        try:
            return await mongo.post(request, obj, collection, default_content)
        except DuplicateKeyError as e:
            raise response.status_conflict(f'Object {obj.display_name} already exists')

async def post_dict(request: Request, obj_dict: DesktopObjectDict, collection: str, type_: type[DesktopObject],
                    default_content: IO | None = None, volume_id: str | None = None,
                    resource_base: str | None = None) -> Response:
    """
    Posts a desktop object dict.

    :param request: the HTTP request (required). The owner property in the obj_dict parameter and the request's
    OIDC_CLAIM_sub header must match.
    :param obj_dict: a desktop object dict (required).
    :param collection: the Mongo collection name (required).
    :param type_: The HEA object type. Required. The type of the object in the request body is compared to this type
    for whether the object is an instance of the given type, and a response with a 400 status code is returned if it is
    not.
    :param default_content: an optional blank document or other default content as a file-like object. This must be not-None
    for any microservices that manage content.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :param resource_base: the base path of the created resource. If None, the collection is used as the base path.
    :return: a Response object with a status of Created and the object's URI in the Location header if the request was
    successful, otherwise a Response object with an error status code.
    """
    try:
        result = await post_dict_return_id(request, obj_dict, collection, type_, default_content, volume_id)
        return await response.post(request, result, resource_base if resource_base is not None else collection)
    except HTTPError as e:
        return e

async def put(request: Request, collection: str, type_: type[DesktopObjectTypeVar], volume_id: str | None = None,
              obj: DesktopObjectTypeVar | None = None,
              pre_save_hook: Callable[[Request, DesktopObjectTypeVar], Awaitable[None]] | None = None) -> Response:
    """
    Updates the HEA object with the specified id.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :param obj: desktop object to use instead of what is in the request body (optional). There will be no validation
    performed on the PUT body.
    :param pre_save_hook: an optional hook to call after the save request has been validated and before saving the
    object.
    :return: a Response object with a status of No Content, Forbidden, or Not Found.
    """
    async with MongoContext(request, volume_id) as mongo:
        if obj is not None:
            obj_ = obj
        else:
            try:
                obj_ = await new_heaobject_from_type(request, type_)
            except DeserializeException as e:
                return response.status_bad_request(str(e))
        try:
            context = mongo.get_default_permission_context(request)
            permitted = await RESTPermissionGroup.PUTTER_PERMS.has_any(obj_, context)
            if not permitted:
                if await RESTPermissionGroup.GETTER_PERMS.has_any(obj_, context):
                    return response.status_forbidden()
                else:
                    return response.status_not_found()
            if pre_save_hook is not None:
                await pre_save_hook(request, obj_)
            result = await mongo.put(request, obj_, collection, context=context)  # if lacks permissions or object is not in database, then updates no records.
        except WriteError as e:
            err_msg = e.details.get('errmsg') if e.details else None
            if e.code == 66:
                return response.status_bad_request(err_msg)
            else:
                return response.status_internal_error(err_msg)
        if result is not None and result.matched_count:
            to_delete = []
            for cache_key in request.app[HEA_CACHE]:
                if cache_key[1] == collection and cache_key[2] in (None, f"id^{request.match_info['id']}"):
                    to_delete.append(cache_key)
            for cache_key in to_delete:
                request.app[HEA_CACHE].pop(cache_key, None)
        return await response.put(bool(result.matched_count if result is not None else None))

async def upsert(request: Request, collection: str, type_: type[DesktopObject], volume_id: str | None = None, filter: Mapping[str, Any] | None = None) -> Response:
    """
    Updates the HEA object, using the specified filter if provided otherwise the object's id, and inserting a new object
    if none matches the filter or the id.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :param filter: optional filter criteria.
    :return: a Response object with a status of No Content, Forbidden, or Not Found.
    """
    sub = request.headers.get(SUB, NONE_USER)
    async with MongoContext(request, volume_id) as mongo:
        try:
            obj = await new_heaobject_from_type(request, type_)
        except DeserializeException as e:
            return response.status_bad_request(str(e).encode())
        try:
            context = mongo.get_default_permission_context(request)
            permitted = await RESTPermissionGroup.PUTTER_PERMS.has_any(obj, context) and \
                await RESTPermissionGroup.POSTER_PERMS.has_any(obj, context)
            if not permitted:
                if await RESTPermissionGroup.GETTER_PERMS.has_any(obj, context):
                    return response.status_forbidden()
                else:
                    return response.status_not_found()

            result = await mongo.upsert_admin(obj, collection, mongoattributes=filter)
        except WriteError as e:
            err_msg = e.details.get('errmsg') if e.details is not None else None
            if e.code == 66:
                return response.status_bad_request(err_msg)
            else:
                return response.status_internal_error(err_msg)
        to_delete = []
        for cache_key in request.app[HEA_CACHE]:
            if cache_key[1] == collection and cache_key[2] in (None, f"id^{request.match_info['id']}"):
                to_delete.append(cache_key)
        for cache_key in to_delete:
            request.app[HEA_CACHE].pop(cache_key, None)
        return await response.put(bool(result))


async def put_content(request: Request, collection: str, type_: type[DesktopObject], volume_id: str | None = None) -> Response:
    """
    Updates the HEA object's associated content.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: a Response object with a status of No Content, Forbidden, or Not Found.
    """
    async with MongoContext(request, volume_id) as mongo:
        context = mongo.get_default_permission_context(request)
        result = await mongo.get(request, collection, var_parts='id', context=context)

        if result is not None:
            obj = heaobject.root.desktop_object_from_dict(result)
            assert obj.id is not None, 'Object id is None'
            permitted = await RESTPermissionGroup.PUTTER_PERMS.has_any(obj, context)
            if not permitted:
                if await RESTPermissionGroup.GETTER_PERMS.has_any(obj, context):
                    return response.status_forbidden()
                else:
                    return response.status_not_found()
            result2 = await mongo.put_content(request, collection, obj.id, obj.display_name)  # if lacks permissions, then updates no records.
            return await response.put(result2)
        else:
            return response.status_not_found()


async def delete(request: Request, collection: str, volume_id: str | None = None) -> Response:
    """
    Deletes the HEA object with the specified id and any associated content.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: No Content, Forbidden, or Not Found.
    """
    async with MongoContext(request, volume_id) as mongo:
        context = mongo.get_default_permission_context(request)
        obj = await mongo.get(request, collection, var_parts='id', context=context)

        if obj is not None:
            obj_ = heaobject.root.desktop_object_from_dict(obj)
            permitted = await RESTPermissionGroup.DELETER_PERMS.has_any(obj_, context)
            if not permitted:
                return response.status_forbidden()
            delete_result = await mongo.delete(request, collection, var_parts='id',
                                               context=context)  # if lacks permissions, then deletes no records.
            if delete_result and delete_result.deleted_count:
                to_delete = []
                for cache_key in request.app[HEA_CACHE]:
                    if cache_key[1] == collection and cache_key[2] in (None, f"id^{request.match_info['id']}"):
                        to_delete.append(cache_key)
                for cache_key in to_delete:
                    request.app[HEA_CACHE].pop(cache_key, None)
            return await response.delete(bool(delete_result.deleted_count) if delete_result else False)
        else:
            return response.status_not_found()



async def ping(request: Request) -> Response:
    """
    Sends a ping command to the database.

    :param request: the HTTP request.
    :return: an HTTP response with status code 200 if the ping is successful, 500 otherwise.
    """
    async with MongoContext(request) as mongo:
        try:
            await mongo.ping()
            return response.status_ok()
        except Exception as e:  # The exact exception is not documented.
            raise response.status_internal_error() from e

async def aggregate(request: Request, collection: str,
                    pipeline: Sequence[Mapping[str, Any]], volume_id: str | None = None) -> Response:
    """
    Execute an aggregation pipeline that returns a list of desktop object dicts that the current user has access to.
    Unlike the other methods that return a response with desktop object dictionaries, this one does not ignore id
    properties returned from mongo, nor does it assume there is an _id ObjectId property and convert it to a str id
    property.

    :param request: the HTTP request (required).
    :param collection. The Mongo collection name (required).
    :param volume_id: the volume_id of the Mongo database. If None, the root volume is assumed.
    :return: a 200 status code response with the desktop object dicts from the pipeline.
    """
    async with MongoContext(request, volume_id) as mongo:
        result: list[DesktopObject] = []
        permissions: list[list[Permission]] = []
        attribute_permissions: list[dict[str, list[Permission]]] = []
        agg = mongo.aggregate(collection, pipeline)
        try:
            context = mongo.get_default_permission_context(request)
            async for r in agg:
                obj = heaobject.root.desktop_object_from_dict(r)
                permitted = await RESTPermissionGroup.GETTER_PERMS.has_any(obj, context)
                if permitted:
                    perms_to_add, attr_perms_to_add = await gather(obj.get_permissions(context), obj.get_all_attribute_permissions(context))
                    permissions.append(perms_to_add)
                    attribute_permissions.append(attr_perms_to_add)
                    result.append(obj)
            return await response.get_all(request, [obj.to_dict() for obj in result], permissions=permissions,
                                          attribute_permissions=attribute_permissions)
        finally:
            await agg.aclose()


async def get_all_desktop_objects(request: Request,
                      collection: str,
                      type_: type[DesktopObject] | None = None,
                      volume_id: str | None = None,
                      mongoattributes: Any | None = None,
                      sort: SortDict | None = None,
                      mongo: Mongo | None = None,
                      context: heaobject.root.PermissionContext | None = None) -> list[DesktopObject]:
    """
    Gets a list of all desktop objects matching the query. This function does not cache its results, but if there are
    cached results from a prior get_all function call, it will use them. The cache key it looks for is
    (sub, collection, None, tuple of sort key-value tuples)

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: the expected type of desktop object in the generator.
    :param volume_id: the id string of the volume containing the requested desktop object. If None, the root volume is
    assumed.
    :param mongoattributes: the mongo query to use. If None, all desktop objects from the collection are returned.
    :param sort: the sort order of the desktop objects as a dict of attributes to 1 (ascending) or -1 (descending).
    :param mongo: the Mongo context. If None, a new Mongo context is created.
    :param context: the PermissionContext. If None, a new context is created.
    :return: a list of DesktopObject.
    """
    sub = request.headers.get(SUB, NONE_USER)
    if mongoattributes is None and not request.query:
        cache_key = (sub, collection, None, tuple((key, val) for key, val in (sort or {}).items()))
        cached_value = request.app[HEA_CACHE].get(cache_key)
    else:
        cached_value = None
    if mongoattributes is None and cached_value is not None:
        return [desktop_object_from_dict(obj_dict, type_=type_) for obj_dict in cached_value[0]]
    else:
        return [obj async for obj in get_all_desktop_objects_gen(request, collection, type_=type_, volume_id=volume_id,
                                                                mongoattributes=mongoattributes, sort=sort,
                                                                mongo=mongo, context=context)]


@overload
def get_all_desktop_objects_gen(request: Request, collection: str, *,
                      volume_id: str | None = None,
                      mongoattributes: Any | None = None,
                      sort: SortDict | None = None,
                      mongo: Mongo | None = None,
                      context: heaobject.root.PermissionContext | None = None) -> AsyncGenerator[DesktopObject, None]: ...


@overload
def get_all_desktop_objects_gen(request: Request, collection: str, *,
                      type_: type[DesktopObjectTypeVar],
                      volume_id: str | None = None,
                      mongoattributes: Any | None = None,
                      sort: SortDict | None = None,
                      mongo: Mongo | None = None,
                      context: heaobject.root.PermissionContext | None = None) -> AsyncGenerator[DesktopObjectTypeVar, None]: ...


@overload
def get_all_desktop_objects_gen(request: Request, collection: str, *,
                      type_: type[DesktopObject] | None,
                      volume_id: str | None = None,
                      mongoattributes: Any | None = None,
                      sort: SortDict | None = None,
                      mongo: Mongo | None = None,
                      context: heaobject.root.PermissionContext | None = None) -> AsyncGenerator[DesktopObject, None]: ...


async def get_all_desktop_objects_gen(request: Request, collection: str, *,
                      type_: type[DesktopObject] | None = None,
                      volume_id: str | None = None,
                      mongoattributes: Any | None = None,
                      sort: SortDict | None = None,
                      mongo: Mongo | None = None,
                      context: heaobject.root.PermissionContext | None = None) -> AsyncGenerator[DesktopObject, None]:
    """
    Gets an async generator of all desktop objects. This generator does no desktop object caching, nor does it use any
    previously cached data.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: the expected type of desktop object in the generator.
    :param volume_id: the id string of the volume containing the requested desktop object. If None, the root volume is
    assumed.
    :param mongoattributes: the mongo query to use. If None, all desktop objects from the collection are returned.
    :param sort: the sort order of the desktop objects as a dict of attributes to 1 (ascending) or -1 (descending).
    :param mongo: the Mongo context. If None, a new Mongo context is created.
    :param context: the PermissionContext. If None, a new context is created.
    :return: an async generator of DesktopObject.
    """
    async def fetch_objects(mongo: Mongo) -> AsyncGenerator[DesktopObject, None]:
        context_ = mongo.get_default_permission_context(request) if context is None else context
        gen = mongo.get_all(request, collection, mongoattributes=mongoattributes, sort=sort,
                            context=context_)
        try:
            async for desktop_object_dict in gen:
                obj = desktop_object_from_dict(desktop_object_dict, type_=type_)
                if await RESTPermissionGroup.GETTER_PERMS.has_any(obj, context_):
                    yield obj
        finally:
            await gen.aclose()

    if mongo:
        f = fetch_objects(mongo)
        try:
            async for obj in f:
                yield obj
        finally:
            await f.aclose()
    else:
        async with MongoContext(request, volume_id) as mongo:
            f = fetch_objects(mongo)
            try:
                async for obj in f:
                    yield obj
            finally:
                await f.aclose()


class MongoSortOrder(Enum):
    """
    Enum for mongo query sort order.
    """
    _value_: Literal[-1, 1]
    ASCENDING = 1
    DESCENDING = -1

    def as_sort_order(self) -> SortOrder | None:
        """
        Converts a MongoSortOrder to a SortOrder.

        :return: a SortOrder.
        """
        return SortOrder.ASC if self is MongoSortOrder.ASCENDING else \
            SortOrder.DESC if self is MongoSortOrder.DESCENDING else None

    def to_sort_dict(self, sort_attribute: str) -> SortDict:
        """
        Returns a sort dict suitable to pass into MongoDB queries.

        :param sort_attributes: an attribute to sort on. Required.
        :return: a dict of attribute name to 1 (ascending) or -1 (descending).
        """
        return {sort_attribute: self.value}

    @overload
    @classmethod
    def from_sort_order(cls, sort_order: SortOrder) -> 'MongoSortOrder': ...

    @overload
    @classmethod
    def from_sort_order(cls, sort_order: None) -> None: ...

    @classmethod
    def from_sort_order(cls, sort_order: SortOrder | None) -> Optional['MongoSortOrder']:
        """
        Converts a SortOrder to a MongoSortOrder.

        :param sort_order: the SortOrder to convert.
        :return: a MongoSortOrder or None if the sort_order is None.
        """
        if sort_order is None:
            return None
        elif sort_order is SortOrder.ASC:
            return cls.ASCENDING
        elif sort_order is SortOrder.DESC:
            return cls.DESCENDING
        else:
            raise ValueError(f'Invalid sort order: {sort_order}')

    @staticmethod
    def from_request(request: Request, default_attr: str | None = None,
                     default_sort_order: Optional['MongoSortOrder'] = None) -> Iterator[tuple[str, 'MongoSortOrder']]:
        """
        Extracts MongoSortOrders by attribute from the request's sort query parameter.

        :param request: the HTTP request. Required.
        :param default_attr: the default attribute to use if no sort attribute is present. If omitted, there must be at
        least as many sort attributes as sort orders.
        :param default_sort_order: the default sort order to use if no sort order is present. If omitted, the default
        sort order is ascending.
        :return: iterator of attribute-MongoSortOrder tuples, if present.
        """
        sort_orders = list(MongoSortOrder.from_sort_order(sort) for sort in extract_sorts(request))
        len_sort_orders = len(sort_orders)
        sort_attrs = list(sort_attr for sort_attr in extract_sort_attrs(request))
        len_sort_attrs = len(sort_attrs)
        max_len = max(len_sort_orders, len_sort_attrs)
        if len_sort_attrs < max_len - 1:
            raise ValueError('Ambiguous sort attributes: more than one sort order specified without an attribute')
        sort_orders_chain = chain(sort_orders, repeat(default_sort_order or MongoSortOrder.ASCENDING, max_len - len_sort_orders))
        if not default_attr:
            return zip(sort_attrs, sort_orders_chain, strict=True)
        else:
            sort_attrs_chain = chain(sort_attrs, repeat(default_attr, max_len - len_sort_attrs))
            return zip(sort_attrs_chain, sort_orders_chain, strict=True)

    @staticmethod
    def from_request_dict(request: Request, default_attr: str | None = None,
                          default_sort_order: Optional['MongoSortOrder'] = None) -> SortDict:
        """
        Extracts the sort order from the request query parameters, and returns a sort dict suitable to pass into
        MongoDB queries.

        :param request: the HTTP request. Required.
        :param default_attr: the default attribute to use if no sort attribute is present. If omitted, there must be at
        least as many sort attributes as sort orders.
        :param default_sort_order: the default sort order to use if no sort order is present. If omitted, the default
        sort order is ascending.
        :return: a dict of attribute names to 1 (ascending) or -1 (descending), or the empty dict if no order is
        specified.
        """
        return {attr_: order.value for attr_, order in MongoSortOrder.from_request(request, default_attr,
                                                                                   default_sort_order=default_sort_order)}

    @staticmethod
    def from_request_dict_raises_http_error(request: Request, default_attr: str | None = None,
                                            default_sort_order: Optional['MongoSortOrder'] = None) -> SortDict:
        """
        Extracts the sort order from the request query parameters, and returns a sort dict suitable to pass into
        MongoDB queries. Raises an HTTPError if the sort query parameter is invalid.

        :param request: the HTTP request. Required.
        :param default_attr: the default attribute to use if no sort attribute is present. If omitted, there must be at
        least as many sort attributes as sort orders.
        :param default_sort_order: the default sort order to use if no sort order is present. If omitted, the default
        sort order is ascending.
        :return: a dict of attribute names to 1 (ascending) or -1 (descending), or the empty dict if no order is
        specified.
        :raises HTTPError: if the sort parameter is invalid.
        """
        try:
            return MongoSortOrder.from_request_dict(request, default_attr=default_attr,
                                                    default_sort_order=default_sort_order)
        except ValueError as e:
            raise response.status_bad_request(f'Invalid sort parameter: {e}')


@overload
def sort_dict_to_cache_key(sort: None) -> None: ...


@overload
def sort_dict_to_cache_key(sort: SortDict) -> SortDictCacheKey: ...


def sort_dict_to_cache_key(sort: SortDict | None) -> SortDictCacheKey | None:
    """
    Converts a sort dict to a cache key tuple.

    :param sort: the sort dict. Required.
    :return: a tuple of the sort keys and values.
    """
    return tuple((key, val) for key, val in sort.items()) if sort is not None else None


@overload
def _dict_to_desktop_object(obj_dict: None, *, type_: type[DesktopObject], obj: DesktopObject | None = None) -> None:
    ...


@overload
def _dict_to_desktop_object(obj_dict: DesktopObjectDict, *, obj: DesktopObject | None = None) -> DesktopObject:
    ...


@overload
def _dict_to_desktop_object(obj_dict: DesktopObjectDict, *, type_: type[DesktopObjectTypeVar],
                            obj: DesktopObjectTypeVar | None = None) -> DesktopObjectTypeVar:
    ...


def _dict_to_desktop_object(obj_dict: DesktopObjectDict | None, *, type_ = DesktopObject, obj:
                            DesktopObject | None = None) -> DesktopObject | None:
    """
    Converts a desktop object dict to a desktop object. If the dictionary is None, None is returned.
    If the dictionary is not None, the object is created from the dictionary and returned.
    If the object is not None, the dictionary is used to update the object and the object is returned.

    :param obj_dict: the dictionary to convert.
    :param type_: the type of the object to create. If None, the type key of the dictionary is used to determine the
    type of the object to create.
    :param obj: the object to populate from the desktop object dict. If None, a new object is created.
    :return: the object created from the dictionary or None if the dictionary is None.
    """
    if obj_dict is None:
        return None
    elif obj:
        if not isinstance(obj, type_):
            raise TypeError(f'Object {obj} is not of type {type_}')
        obj.from_dict(obj_dict)
        return obj
    else:
        return heaobject.root.desktop_object_from_dict(obj_dict, type_=type_)



from unittest import IsolatedAsyncioTestCase
from heaserver.service.backgroundtasks import BackgroundTasks
from asyncio import create_task, sleep, CancelledError


class BackgroundTasksTestCase(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.__background_tasks = BackgroundTasks()

    async def asyncTearDown(self) -> None:
        await self.__background_tasks.clear()

    async def test_contains(self):
        async def coro(app):
            pass

        await self.__background_tasks.add(coro, 'test_contains')
        self.assertTrue(self.__background_tasks.contains('test_contains'))

    async def test_does_not_contain(self):
        async def coro(app):
            pass

        await self.__background_tasks.add(coro, 'test_contains')
        self.assertTrue(not self.__background_tasks.contains('test_contains2'))

    async def test_cancel(self):
        async def coro(app):
            while True:
                sleep(0.1)

        await self.__background_tasks.add(coro, 'test_cancel')
        self.__background_tasks.cancel('test_cancel')
        await self.__background_tasks.join_all()
        self.assertTrue(self.__background_tasks.done('test_cancel'))

    async def test_cancel_all(self):
        async def coro1(app):
            while True:
                sleep(0.1)

        async def coro2(app):
            while True:
                sleep(0.1)

        await self.__background_tasks.add(coro1, 'test_cancel1')
        await self.__background_tasks.add(coro2, 'test_cancel2')
        self.__background_tasks.cancel_all()
        await self.__background_tasks.join_all()
        self.assertTrue(self.__background_tasks.done('test_cancel1') and self.__background_tasks.done('test_cancel2'))

    async def test_done(self):
        async def coro(app):
            pass

        await self.__background_tasks.add(coro, 'test_is_done')
        await self.__background_tasks.join('test_is_done')
        self.assertTrue(self.__background_tasks.done('test_is_done'))

    async def test_succeeded(self):
        async def coro(app):
            pass

        await self.__background_tasks.add(coro, 'test_succeeded')
        await self.__background_tasks.join('test_succeeded')
        self.assertTrue(self.__background_tasks.succeeded('test_succeeded'))

    async def test_failed(self):
        async def coro(app):
            raise ValueError

        await self.__background_tasks.add(coro, 'test_failed')
        await self.__background_tasks.join('test_failed')
        self.assertTrue(self.__background_tasks.failed('test_failed'))

    async def test_get_error(self):
        async def coro(app):
            raise ValueError

        await self.__background_tasks.add(coro, 'test_get_error')
        await self.__background_tasks.join('test_get_error')
        self.assertIsInstance(self.__background_tasks.error('test_get_error'), ValueError)

    async def test_remove(self):
        async def coro(app):
            pass

        await self.__background_tasks.add(coro, 'test_remove')
        await self.__background_tasks.join('test_remove')
        self.__background_tasks.remove('test_remove')
        self.assertFalse(self.__background_tasks.contains('test_remove'))

    async def test_len(self):
        async def coro1(app):
            pass

        async def coro2(app):
            pass

        await self.__background_tasks.add(coro1, 'test_len1')
        await self.__background_tasks.add(coro2, 'test_len2')
        await self.__background_tasks.join('test_len1')
        try:
            self.assertEqual(2, len(self.__background_tasks))
        finally:
            await self.__background_tasks.join_all()

    async def test_clear(self):
        async def coro1(app):
            pass

        async def coro2(app):
            pass

        await self.__background_tasks.add(coro1, 'test_clear1')
        await self.__background_tasks.add(coro2, 'test_clear2')
        await self.__background_tasks.join_all()
        await self.__background_tasks.clear()
        self.assertEqual(0, len(self.__background_tasks))

    async def test_auto_join(self):
        async def coro1(app):
            pass

        async def coro2(app):
            pass

        await self.__background_tasks.add(coro1, 'test_clear1')
        await self.__background_tasks.add(coro2, 'test_clear2')
        task = create_task(self.__background_tasks.auto_join())
        while self.__background_tasks.running_tasks() > 0:
            await sleep(0.1)
        task.cancel()
        try:
            await task
        except CancelledError:
            pass
        self.assertTrue(self.__background_tasks.done('test_clear1') and self.__background_tasks.done('test_clear2'))

    async def test_done_tasks(self):
        async def coro1(app):
            pass

        async def coro2(app):
            pass

        await self.__background_tasks.add(coro1, 'test_done_tasks1')
        await self.__background_tasks.add(coro2, 'test_done_tasks2')
        await self.__background_tasks.join_all()
        self.assertEqual(2, self.__background_tasks.done_tasks())

    async def test_auto_join_with_error(self):
        async def coro1(app):
            raise ValueError

        async def coro2(app):
            pass

        await self.__background_tasks.add(coro1, 'test_clear1')
        await self.__background_tasks.add(coro2, 'test_clear2')
        task = create_task(self.__background_tasks.auto_join())
        while self.__background_tasks.running_tasks() > 0:
            await sleep(0.1)
        task.cancel()
        try:
            await task
        except CancelledError:
            pass
        self.assertTrue(
            self.__background_tasks.failed('test_clear1') and self.__background_tasks.succeeded('test_clear2'))



from unittest import TestCase
from aiohttp.test_utils import make_mocked_request
from heaserver.service.db.mongoservicelib import MongoSortOrder


class MongoServiceLibTestCase(TestCase):

    def test_sort_from_request(self):
        """Make sure one sort order without an attribute correctly uses the provided default attribute."""
        request = make_mocked_request('GET', '/test?sort=asc')
        self.assertEqual([('foo', MongoSortOrder.ASCENDING)], list(MongoSortOrder.from_request(request, 'foo')))

    def test_sort_from_request_key_case_sensitivity(self):
        """Make sure one sort order without an attribute correctly uses the provided default attribute."""
        request = make_mocked_request('GET', '/test?SORT=asc')
        self.assertEqual([], list(MongoSortOrder.from_request(request, 'foo')))

    def test_sort_from_request_value_case_insensitivity(self):
        """Make sure one sort order without an attribute correctly uses the provided default attribute."""
        request = make_mocked_request('GET', '/test?sort=ASC')
        self.assertEqual([('foo', MongoSortOrder.ASCENDING)], list(MongoSortOrder.from_request(request, 'foo')))

    def test_sort_multiple_from_request(self):
        """Make sure multiple sort orders without attributes correctly error out."""
        request = make_mocked_request('GET', '/test?sort=asc&sort=desc')
        with self.assertRaises(ValueError):
            self.assertEqual([('foo', MongoSortOrder.ASCENDING)],
                            list(MongoSortOrder.from_request(request, 'foo')))

    def test_sort_two_sorts_and_one_attr_from_request(self):
        """Make sure one missing sort attribute is replaced with the provided default."""
        request = make_mocked_request('GET', '/test?sort=asc&sort=desc&sort_attr=bar')
        self.assertEqual([('bar', MongoSortOrder.ASCENDING),
                          ('foo', MongoSortOrder.DESCENDING)],
                          list(MongoSortOrder.from_request(request, 'foo')))

    def test_sort_two_orders_and_two_attrs_from_request(self):
        """Make sure sort attributes and orders are correctly extracted."""
        request = make_mocked_request('GET', '/test?sort=asc&sort=desc&sort_attr=bar&sort_attr=foo')
        self.assertEqual([('bar', MongoSortOrder.ASCENDING),
                          ('foo', MongoSortOrder.DESCENDING)],
                          list(MongoSortOrder.from_request(request)))

    def test_sort_two_orders_and_three_attrs_from_request(self):
        """Make sure sort attributes without orders are sorted in ascending order."""
        request = make_mocked_request('GET', '/test?sort=asc&sort=desc&sort_attr=bar&sort_attr=foo&sort_attr=baz')
        self.assertEqual([('bar', MongoSortOrder.ASCENDING),
                          ('foo', MongoSortOrder.DESCENDING),
                          ('baz', MongoSortOrder.ASCENDING)],
                          list(MongoSortOrder.from_request(request)))

    def test_sort_three_attrs_and_no_orders_from_request(self):
        """Make sure sort attributes without orders are sorted in ascending order."""
        request = make_mocked_request('GET', '/test?sort_attr=bar&sort_attr=foo&sort_attr=baz')
        self.assertEqual([('bar', MongoSortOrder.ASCENDING),
                          ('foo', MongoSortOrder.ASCENDING),
                          ('baz', MongoSortOrder.ASCENDING)],
                          list(MongoSortOrder.from_request(request)))

    def test_sort_as_dict(self):
        """Make sure sort orders and attributes are correctly extracted as a dict."""
        request = make_mocked_request('GET', '/test?sort=asc&sort=desc&sort_attr=bar&sort_attr=foo')
        self.assertEqual({'bar': MongoSortOrder.ASCENDING.value,
                          'foo': MongoSortOrder.DESCENDING.value},
                          MongoSortOrder.from_request_dict(request))



import asyncio
import configparser
import logging
from typing import Optional, TypeVar, Any, AsyncGenerator

import opensearchpy
from aiohttp import web
from cachetools import TTLCache
from heaobject.error import DeserializeException
from heaobject.keychain import Credentials
from heaobject.root import HEAObjectDict, json_dumps
from heaobject.volume import OpenSearchFileSystem, FileSystem
from opensearchpy import AsyncOpenSearch, RequestError, NotFoundError
from heaserver.service.aiohttp import extract_sub
from heaserver.service.appproperty import HEA_DB
from heaserver.service.db.aws import S3
from heaserver.service.db.database import Database, DatabaseContextManager, MicroserviceDatabaseManager, \
    get_file_system_and_credentials_from_volume, get_credentials_from_volume
from configparser import ConfigParser
from heaobject.folder import Item

from heaserver.service.db.mongo import Mongo

from heaserver.service.testcase.util import run_coroutine_in_new_event_loop

CONFIG_SECTION="Opensearch"

ItemTypeVar = TypeVar('ItemTypeVar', bound=Item)

class OpenSearch(Database):

    def __init__(self, config: Optional[ConfigParser],
                 host: str | None = None,
                 port: int | None = None,
                 use_ssl: bool | None = False,
                 verify_certs: bool | None = False,
                 index: str | None = None,
                 username: str | None = None,
                 password: str | None = None,
                 **kwargs):
        """
        Initializes the OpenSearch client and stores configuration parameters.

        :param host: host of opensearch microservice
        :param port: port of opensearch microservice
        :param use_ssl: Whether to use SSL for the connection.
        :param verify_certs: Whether to verify SSL certificates.
        :param index_name: The name of the index to act upon.
        """
        super().__init__(config, **kwargs)
        self.__host = host
        self.__port = port
        self.__use_ssl = use_ssl
        self.__verify_certs = verify_certs
        self.__index = index
        self.__scroll_ids: list[str] = []  # Keep track of scroll IDs
        self.__volume_id_to_credentials: TTLCache[tuple[str, str], Credentials | None] = TTLCache(maxsize=128, ttl=30)
        self.__username = username
        self.__password = password

        if config and CONFIG_SECTION in config:
            _section = config[CONFIG_SECTION]
            self.__host = _section.get('Hostname', fallback=self.__host)
            self.__port = _section.getint( 'Port', fallback=self.__port)
            self.__use_ssl = _section.getboolean('UseSSL', fallback=self.__use_ssl)
            self.__verify_certs = _section.getboolean('VerifyCerts', fallback=self.__verify_certs)
            self.__index = _section.get('Index', fallback=self.__index)
            self.__username = _section.get('Username', fallback=self.__username)
            self.__password = _section.get('Password', fallback=self.__password)

        self.__client = AsyncOpenSearch(
            hosts=[{'host': self.__host, 'port': self.__port}],
            http_auth=(self.__username, self.__password) if self.__username and self.__password else None,
            use_ssl=self.__use_ssl,
            verify_certs=self.__verify_certs,
        )

    @property
    def file_system_type(self) -> type[FileSystem]:
        return OpenSearchFileSystem

    async def get_credentials_from_volume(self, request: web.Request, volume_id: str) -> Credentials | None:
        sub = extract_sub(request)
        if (credentials := self.__volume_id_to_credentials.get((sub, volume_id))) is not None:
            return credentials
        else:
            # mypy can't seem to distinguish between this method and the module-level function.
            credentials = await get_credentials_from_volume(request, volume_id, Credentials)  # type:ignore[func-returns-value]
            self.__volume_id_to_credentials[(sub, volume_id)] = credentials
            return credentials

    def __filter_doc_fields(self, doc: HEAObjectDict, mapping: dict) -> dict:
        """
        Filters a document to include only fields that exist in the index mapping.

        :param doc: The original document to filter.
        :param mapping: The mapping of the index.
        :return: A new document containing only the fields that are present in the mapping.
        """

        properties = mapping.get("properties", {})
        filtered_doc: dict[str, Any] = {}

        for key, value in doc.items():
            if key in properties:
                field_type = properties[key].get("type")
                if field_type == "text" and isinstance(value, str):
                    filtered_doc[key] = value
                elif field_type == "keyword" and isinstance(value, str):
                    filtered_doc[key] = value
                elif field_type == "integer" and isinstance(value, int):
                    filtered_doc[key] = value
                elif field_type == "boolean" and isinstance(value, bool):
                    filtered_doc[key] = value
                elif field_type == "float" and isinstance(value, (int, float)):
                    filtered_doc[key] = float(value)
                elif value is None:
                    filtered_doc[key] = None
                # Add more type checks if needed for specific field types
        return filtered_doc

    async def _get_index_mapping(self, index: str | None) -> dict :
        """
        Retrieves the mapping of the specified OpenSearch index.

        :param index: The name of the index.
        :return: The mapping of the index as a dictionary.
        """
        try:
            response = await self.__client.indices.get_mapping(index=index)
        except opensearchpy.TransportError as te:
            raise te
        return response.get(index, {}).get("mappings", {})

    async def create_index(self, index: Optional[str] = None, body: Optional[dict] = None) -> bool:
        """
        Creates an OpenSearch index.

        :param index: The name of the index to create. Defaults to the class variable __index.
        :param body: (Optional) A dictionary specifying the index settings and mappings.
        :return: True if the index was created successfully, False otherwise.
        :raises ValueError: If no index name is provided or the class variable __index is not defined.
        """
        if not index and not self.__index:
            raise ValueError("Index name must be provided.")
        idx = index if index else self.__index

        try:
            await self.__client.indices.create(index=idx, body=body)
            return True
        except opensearchpy.RequestError as e:
            if e.error == "resource_already_exists_exception":
                logging.warning(f"Index '{idx}' already exists.")
            else:
                raise e
        return False


    async def delete_index(self, index: str | None = None) -> bool:
        """
        Deletes an OpenSearch index.

        :param index: The name of the index to delete. Defaults to the class variable __index.
        :return: True if the index was deleted successfully, False otherwise.
        :raises ValueError: If no index name is provided or the class variable __index is not defined.
        """
        if not index and not self.__index:
            raise ValueError("Index name must be provided.")
        idx = index if index else self.__index

        try:
            await self.__client.indices.delete(index=idx)
            return True
        except opensearchpy.NotFoundError:
            logging.warning(f"Index '{idx}' does not exist.")
        return False

    async def insert(self, doc: HEAObjectDict, index: str | None = None) -> Optional[str]:
        """
        Updates or inserts a document in the specified OpenSearch index.

        :param doc: The document data to be inserted or updated. If the document contains an `id` field,
                    it will be used as the document ID. Otherwise, a new ID will be generated.
        :param index: (Optional) The name of the OpenSearch index where the document should be updated.
                      If not provided, the default index will be used.

        :return: The ID of the updated or inserted document, or `None` if the operation failed.

        :raises ValueError: Raised if the `index` parameter is not provided.
        :raises Exception: Raised for any other errors encountered during the update operation.
        """
        logger = logging.getLogger(__name__)
        try:
            if not index and not self.__index:
                raise ValueError("Index name must be provided.")
            idx = index if index else self.__index
            # Fetch index mapping
            mapping = await self._get_index_mapping(idx)

            # Filter document fields based on the mapping
            logger.debug(f"mapping: {mapping}" )
            doc_id = doc.get('id', None)
            filtered_doc = self.__filter_doc_fields(doc, mapping)

            response = await self.__client.index(
                index=idx,
                id=doc_id,
                body=filtered_doc,
                refresh=True  # Optional: Refresh the index to make the change immediately searchable
            )
            return response["_id"]
        except Exception as e:
            logger.debug(f"Failed to update document: {e}")
            return None

    async def batch_insert(self, bulk_doc_gen: AsyncGenerator[ItemTypeVar, None]) -> bool:
        """
        Insert multiple documents into the index in bulk.
        :param bulk_doc_gen: An asynchronous generator of documents to be inserted.
        :return: True if all documents were successfully inserted, False otherwise.
        """
        return await self.__batch_operation(bulk_doc_gen, action_type="index")

    async def batch_delete(self, bulk_doc_ids: AsyncGenerator[str, None]) -> bool:
        """
        Delete multiple documents from the index in bulk.
        :param bulk_doc_ids: An asynchronous generator of document IDs to be deleted.
        :return: True if all documents were successfully deleted, False otherwise.
        """
        return await self.__batch_operation(bulk_doc_ids, action_type="delete")

    async def __batch_operation(self, bulk_gen: AsyncGenerator[Any, None], action_type: str) -> bool:
        """
        Perform bulk operations (insert/delete) on the index.
        :param bulk_gen: An asynchronous generator of documents or document IDs.
        :param action_type: The type of bulk operation ('index' for insert, 'delete' for delete).
        :return: True if the operation was successful, False otherwise.
        """
        logger = logging.getLogger(__name__)
        success = False
        try:
            actions = []
            async for item in bulk_gen:
                if action_type == "index":
                    action = {"index": {"_index": self.__index, "_id": item.id}}
                    actions.append(f"{json_dumps(action)}\n{json_dumps(item.to_dict())}")
                elif action_type == "delete":
                    action = {"delete": {"_index": self.__index, "_id": item}}
                    actions.append(json_dumps(action))
                else:
                    raise ValueError(f"Unsupported action type: {action_type}")

                # Send in batches of 1000 for efficiency
                if len(actions) >= 1000:
                    bulk_body = "\n".join(actions) + "\n"
                    response = await self.__client.bulk(body=bulk_body)
                    if response["errors"]:
                        raise ValueError(f"Some documents failed during batch {action_type}.")
                    actions.clear()

            # Process remaining items if any
            if actions:
                bulk_body = "\n".join(actions) + "\n"
                response = await self.__client.bulk(body=bulk_body)
                if response["errors"]:
                    logger.debug("Some documents failed during batch %s." % action_type)
                    return False
            success = True
        except RequestError as re:
            logger.exception("Failed to perform bulk operation: %s", re)
            success = False
        except Exception as e:
            logger.exception("Failed to perform bulk operation: %s", e)
            success = False
        return success

    async def get_doc(self, item_id: str, search_item_type: type[ItemTypeVar], index: str | None = None) -> ItemTypeVar | None:
        """
        Fetch a single item from OpenSearch by its document ID.

        :param item_id: The ID of the document to retrieve.
        :param search_item_type: The type used to deserialize the document.
        :param index: (Optional) the index to query against.
        :return: A single deserialized item or None if not found.
        """
        try:
            if not index and not self.__index:
                raise ValueError("Index name must be provided.")
            idx = index if index else self.__index

            response = await self.__client.get(index=idx, id=item_id)
            if "_source" not in response:
                return None

            item = search_item_type()
            response["_source"]["id"] = response["_id"]
            response["_source"]["type"] = item.get_type_name()
            item.from_dict(response["_source"])
            return item

        except NotFoundError:
            return None  # Let caller handle missing doc case cleanly
        except Exception as e:
            logging.exception(f"Unexpected error getting item {item_id}: {e}")
            raise



    async def search(self, query: dict, search_item_type: type[ItemTypeVar],
                     index: str | None = None,
                     max_results: int = 1000,
                     size: int = 100
                     ) -> list[ItemTypeVar]:
        """
        Perform a paginated search with a maximum limit of results.

        :param search_item_type: The type used to deserialize the documents.
        :param index: The OpenSearch index to query.
        :param query: The query dictionary.
        :param size: Number of results per page.
        :param max_results: The maximum number of results to retrieve.
        :return: A list of documents matching the query.
        """
        search_items = []
        total_results = 0
        from_offset = 0  # Used for pagination

        try:
            while total_results < max_results:
                paginated_query = {
                    **query,
                    "from": from_offset,
                    "size": size,
                    "sort": [{"_doc": {"order": "asc"}}]  # Ensures consistent ordering for pagination
                }

                response = await self.__client.search(index=index, body=paginated_query)
                hits = response.get("hits", {}).get("hits", [])

                if not hits:
                    break  # No more results

                for hit in hits:
                    hit_id = hit["_id"]
                    if "_source" in hit:
                        search_item = search_item_type()
                        hit["_source"]["id"] = hit_id
                        hit["_source"]["type"] = search_item.get_type_name()
                        search_item.from_dict(hit["_source"])
                        search_items.append(search_item)
                        total_results += 1

                        if total_results >= max_results:
                            break  # Stop if max limit reached

                from_offset += size  # Move to next batch

        except DeserializeException as de:
            raise de
        except Exception as e:
            raise e

        return search_items


    async def delete_doc(self, item_id: str, index: str | None = None) -> bool:
        """
        Delete a single document from OpenSearch by its document ID.

        :param item_id: The ID of the document to delete.
        :param index: (Optional) the index to delete from.
        :return: True if the document was deleted, False if not found.
        """
        try:
            if not index and not self.__index:
                raise ValueError("Index name must be provided.")
            idx = index if index else self.__index

            response = await self.__client.delete(index=idx, id=item_id, ignore=[404])
            return response.get("result") == "deleted"

        except Exception as e:
            logging.exception(f"Failed to delete item {item_id}: {e}")
            raise e

    def really_close(self):
        logger = logging.getLogger(__name__)
        if self.__client:
            logger.debug("Closing OpenSearch client...")
            run_coroutine_in_new_event_loop(self.__client.close())
            logger.debug("OpenSearch client closed.")


class S3WithOpenSearch(S3, OpenSearch):
    def __init__(self, config:  configparser.ConfigParser | None = None, **kwargs):
        super().__init__(config, **kwargs)

class MongoWithOpenSearch(Mongo, OpenSearch):
    def __init__(self, config:  configparser.ConfigParser | None = None, **kwargs):
        super().__init__(config, **kwargs)

class S3AndMongoWithOpenSearch(S3, Mongo, OpenSearch):
    def __init__(self, config:  configparser.ConfigParser | None = None, **kwargs):
        super().__init__(config, **kwargs)

class OpenSearchManager(MicroserviceDatabaseManager):

    def __init__(self, config:  configparser.ConfigParser | None = None):
        super().__init__(config)


    def get_database(self) -> OpenSearch:
        """
        Initializes and returns an instance of OpenSearchClient with the provided configuration.
        """
        return OpenSearch(config=self.config, managed=True)

    @classmethod
    def database_types(self) -> list[str]:
        return ['system|opensearch']


class S3WithOpenSearchManager(OpenSearchManager):

    def get_database(self) -> S3WithOpenSearch:
        """
        Initializes and returns an instance of OpenSearchClient with the provided configuration.
        """
        client = S3WithOpenSearch(config =self.config, managed=True)
        return client


class MongoWithOpenSearchManager(OpenSearchManager):

    def get_database(self) -> MongoWithOpenSearch:
        """
        Initializes and returns an instance of OpenSearchClient with the provided configuration.
        """
        client = MongoWithOpenSearch(config=self.config, managed=True)
        return client

class S3MongoOpenSearchMananger(OpenSearchManager):

    def get_database(self) -> S3AndMongoWithOpenSearch:
        """
        Initializes and returns an instance of OpenSearchClient with the provided configuration.
        """
        client = S3AndMongoWithOpenSearch(config =self.config, managed=True)
        return client


class OpenSearchContext(DatabaseContextManager[OpenSearch, Credentials]): # Go into db package?
    """
    Provides a OpenSearch index connection object. If neither a volume nor a credentials object is passed into the
    constructor, the host, port in the microservice's configuration file will be used, it will use defaults of OpenSearch
    filesystem.
    """

    async def connection(self) -> OpenSearch:
        return await _get_opensearch(self.request, self.volume_id)



async def _get_opensearch(request: web.Request, volume_id: Optional[str]) -> OpenSearch:
    """
    Gets a opensearch client.

    :param request: the HTTP request (required).
    :param volume_id: the id string of a volume.
    :return: a OpenSearch client for the file system specified by the volume's file_system_name attribute. If no volume_id
    was provided, the return value will be the "default" OpenSearch client for the microservice found in the HEA_DB
    application-level property.
    :raise ValueError: if there is no volume with the provided volume id, the volume's file system does not exist,
    or a necessary service is not registered.
    """

    if volume_id is not None:
        file_system, credentials = await get_file_system_and_credentials_from_volume(request, volume_id, OpenSearchFileSystem)
        if credentials is None:
            return OpenSearch(None, host=file_system.host, port=file_system.port, index=file_system.index)
        else:
            return OpenSearch(None, host=file_system.host, port=file_system.port, index=file_system.index,
                              username=credentials.account, password=credentials.password)
    else:
        return request.app[HEA_DB]



from unittest import IsolatedAsyncioTestCase


from .awss3foldertestcase import AWSS3FolderTestCase
from heaserver.service.db.aws import S3
import boto3
from moto import mock_aws


class TestS3GetAccount(IsolatedAsyncioTestCase):

    @mock_aws()
    def run(self, result):
        super().run(result)

    async def test_get_account(self):
        sts = boto3.client('sts')
        actual = await S3._get_basic_account_info(sts)
        self.assertEqual('123456789012', actual.id)


class TestAWSProperties(AWSS3FolderTestCase):
    async def test_get_property(self):
        async with self.client.request('GET', '/properties/CLOUD_AWS_CRED_URL') as resp:
            self.assertEqual(200, resp.status)

    async def test_get_property_not_found(self):
        async with self.client.request('GET', '/properties/TEST') as resp:
            self.assertEqual(404, resp.status)


class TestMockAWSAccount(AWSS3FolderTestCase):
    @mock_aws()
    def run(self, result):
        super().run(result)

    async def test_get_volume_id_for_account_id(self):
        async with self.client.request('GET', '/accounts/123456789012') as resp:
            self.assertEqual(200, resp.status)





"""
An expression evaluator, customized for HEA. It's like the built-in eval function, but safe. Only the following names,
operators, and functions will be made available:

Names:
    All HEAObject classes, subclasses, packages, and modules in the heaobject library that do not begin
    with an underscore.
    All attributes of the provided HEAObject, provided by the get_attributes method.

Operators:
    All math, boolean, and equality operators.

Functions:
    is_heaobject(cls: HEAObject): returns whether the object is an instance of the provided class.

Examples:
    >>> from heaobject.folder import AWSS3Folder
    >>> from heaserver.service.expression import get_eval_for
    >>> obj = AWSS3Folder()
    >>> obj.display_name = 'foobar'
    >>> obj.description = 'this is a description'
    >>> eval_ = get_eval_for(obj).eval
    >>> print(eval_('isheaobject(heaobject.folder.AWSS3Folder)'))
    >>> print(eval_('display_name == "foobar"'))
    >>> print(eval_('id is not None'))
    >>> print(eval_('id is None or display_name is None'))
    >>> print(eval_('"this" in description'))

"""
import warnings

from simpleeval import SimpleEval, NameNotDefined, FunctionNotDefined, FeatureNotAvailable, AssignmentAttempted
from typing import Protocol, Any
import heaobject
from heaobject.root import HEAObject, HEAObjectDict, from_dict
from collections.abc import Mapping
import importlib, pkgutil, inspect, heaobject


class EvalProtocol(Protocol):
    """
    A protocol for objects with an eval method accepting one parameter, an expression string.
    """
    def eval(self, expr: str):
        """
        Evaluates the provided expression.

        :param expr: the expression string.
        :return: the value of the expression.
        :raises EvaluatorExpression: if a parsing error occurs.
        """
        ...


class EvaluatorException(ValueError):
    """
    Raised when a parsing error occurs evaluating the expression.
    """
    pass


class Evaluator:
    """
    An expression evaluator, customized for HEA.
    """
    def __init__(self, wrapped: EvalProtocol):
        self.__wrapped = wrapped

    def eval(self, expr: str) -> Any:
        """
        Evaluates the provided expression.

        :param expr: the expression string. Cannot be None or the empty string.
        :return: the value of the expression.
        :raises EvaluatorExpression: if a parsing error occurs.
        """
        if expr is None:
            raise EvaluatorException(f'Expression cannot be None')
        elif not expr:
            raise EvaluatorException(f'Expression cannot be the empty string')
        try:
            with warnings.catch_warnings():
                warnings.simplefilter('error')
                return self.__wrapped.eval(expr)
        except (NameNotDefined, FunctionNotDefined, FeatureNotAvailable) as e:
            raise EvaluatorException(str(e)) from e
        except AssignmentAttempted as e:
            raise EvaluatorException(f'Assignment {expr} attempted')


def get_eval_for(obj: HEAObject | HEAObjectDict, extra_functions: Mapping[str, Any] | None = None, extra_names: Mapping[str, Any] | None = None) -> EvalProtocol:
    """
    Creates an expression evaluator, customized for HEA. Loads the provided HEAObject, as well as any provided extra
    names.

    :param obj: the HEAObject to load into the evaluator (required).
    :param extra_functions: a mapping of function names to functions to make available to the evaluator. It automatically
    adds the isheaobject function, which returns whether the object is an instance of the provided class. Passing a
    function with the same name will override the default isheaobject function.
    :param extra_names: any extra names to make available to the evaluator.
    :return: an object implementing the EvalProtocol protocol.

    Attempts at variable assignment will be ignored, but a warning will be generated. The type of warning is undefined
    beyond being a UserWarning.
    """
    names: dict[str, Any] = {}
    names[heaobject.__name__] = heaobject
    for module in pkgutil.iter_modules(heaobject.__path__, prefix='heaobject.'):
        module_ = importlib.import_module(module.name)
        if not any(module_part.startswith('_') for module_part in module.name.split('.')):
            names[module.name] = module_
        for name, cls in inspect.getmembers(module_):
            if inspect.isclass(cls) and issubclass(cls, HEAObject):
                names[name] = cls

    hea_obj, obj_dict = (obj, obj.to_dict()) if isinstance(obj, HEAObject) else (from_dict(obj), obj)
    names.update(obj_dict)
    if extra_names:
        names.update(extra_names)

    def isheaobject(cls: type[HEAObject]):
        if issubclass(cls, HEAObject):
            return isinstance(hea_obj, cls)
        else:
            raise ValueError(f'cls {cls} is not a HEAObject')

    functions = {'isheaobject': isheaobject}
    functions.update(extra_functions or {})

    return Evaluator(SimpleEval(functions=functions, names=names))




from unittest import TestCase
from heaserver.service.uritemplate import tvars


class URITemplateVarsTestCase(TestCase):
    def test_one_variable(self) -> None:
        """
        Checks if the route /organizations/{id}/members correctly captures an id in the given URL when it is
        present.
        """
        self.assertEqual({'id': 'qwertyuiop'},
                         tvars('/organizations/{id}/members', '/organizations/qwertyuiop/members'))

    def test_one_variable_not_present(self) -> None:
        """
        Checks if the route /organizations/{id}/members correctly does not capture an id in the given URL when it is
        not present.
        """
        self.assertEqual({}, tvars('/organizations/{id}/members', '/organizations/'))

    def test_two_variables_both_present(self) -> None:
        """
        Checks if the route /volumes/{volume_id}/organizations/{id} correctly captures a volume id and an id in the
        given URL when both are present.
        """
        self.assertEqual({'volume_id': 'asdfghjkl', 'id': 'qwertyuiop'},
                         tvars('/volumes/{volume_id}/organizations/{id}',
                               '/volumes/asdfghjkl/organizations/qwertyuiop'))

    def test_two_variables_one_present(self) -> None:
        """
        Checks if the route /volumes/{volume_id}/organizations/{id} correctly only captures a volume in the given URL
        when the volume_id is present but the id is not.
        """
        self.assertEqual({'volume_id': 'asdfghjkl'},
                         tvars('/volumes/{volume_id}/organizations/{id}',
                               '/volumes/asdfghjkl/organizations/'))

    def test_no_variables(self) -> None:
        """Checks if the route /foo/bar/spam correctly captures nothing when the URL is identical."""
        self.assertEqual({}, tvars('/foo/bar/spam', '/foo/bar/spam'))

    def test_decoding_spaces(self) -> None:
        """
        Checks if the route /foo/{var}/spam correctly captures a variable and decodes it when necessary.
        """
        self.assertEqual({'var': 'a value with spaces'},
                         tvars('/foo/{var}/spam', '/foo/a%20value%20with%20spaces/spam'))

    def test_decoding_slashes(self) -> None:
        """
        Checks if the route /foo/{var}/spam correctly captures a variable and decodes it when necessary.
        """
        self.assertEqual({'var': 'a/value/with/slashes'},
                         tvars('/foo/{var}/spam', '/foo/a%2Fvalue%2Fwith%2Fslashes/spam'))

    def test_decoding_pipes(self) -> None:
        """
        Checks if the route /foo/{var}/spam correctly captures a variable and decodes it when necessary.
        """
        self.assertEqual({'var': 'a|value|with|pipes'},
                         tvars('/foo/{var}/spam', '/foo/a%7Cvalue%7Cwith%7Cpipes/spam'))



S3_GET_OBJECT = 's3:GetObject'
S3_PUT_OBJECT = 's3:PutObject'
S3_DELETE_OBJECT = 's3:DeleteObject'
S3_GET_OBJECT_TAGGING = 's3:GetObjectTagging'
S3_PUT_OBJECT_TAGGING = 's3:PutObjectTagging'
S3_LIST_BUCKET = 's3:ListBucket'
S3_CREATE_BUCKET = 's3:CreateBucket'
S3_DELETE_BUCKET = 's3:DeleteBucket'
S3_GET_BUCKET_LOCATION = 's3:GetBucketLocation'  # Permission to get the bucket's region
S3_GET_BUCKET_OBJECT_LOCK_CONFIGURATION = 's3:GetBucketObjectLockConfiguration'
S3_GET_BUCKET_TAGGING = 's3:GetBucketTagging'
S3_PUT_BUCKET_TAGGING = 's3:PutBucketTagging'
S3_GET_BUCKET_VERSIONING = 's3:GetBucketVersioning'
S3_PUT_BUCKET_VERSIONING = 's3:PutBucketVersioning'
S3_LIST_OBJECT_VERSIONS = 's3:ListBucketVersions'
S3_DELETE_OBJECT_VERSION = 's3:DeleteObjectVersion'



import logging

from functools import wraps
from typing import Optional, Type, Any, TypeVar, Generic, AsyncContextManager, cast
from collections.abc import AsyncIterable, Awaitable, Coroutine, Callable, Generator, Iterable, Iterator, Collection, \
    Mapping
from collections import OrderedDict
import requests
from contextlib import contextmanager
import asyncio
import time
import os
from dataclasses import dataclass
from heaobject.util import seconds_since_epoch, now  # Keep here so that older uses of it don't break.
import inspect
from contextlib import asynccontextmanager
from datetime import datetime
from email.utils import formatdate


RT = TypeVar('RT')  # return type

def async_retry(*exceptions: type[Exception], retries: int = 3, cooldown: int | None = 1,
                verbose = True) -> Callable[[Callable[..., Coroutine[Any, Any, RT]]], Callable[..., Coroutine[Any, Any, RT]]]:
    """
    Decorate an async function to execute it a few times before giving up.

    :param exceptions: One or more exceptions expected during function execution, as positional arguments.
    :param retries: Number of retries of function execution.
    :param cooldown: Seconds to wait before retry. If None, does not cool down.
    :param verbose: Specifies if we should log about not successful attempts.
    :raises: if all retries failed, or an exception other than those in the exceptions parameter was raised, the
    exception from the last try.
    """

    def wrap(func):
        @wraps(func)
        async def inner(*args, **kwargs) -> Any:
            _logger = logging.getLogger(__name__)
            retries_count = 0

            while True:
                try:
                    _logger.debug('About to try %s', func)
                    result = await func(*args, **kwargs)
                except exceptions as err:
                    message = "Exception {} during {} execution. " \
                              "{} of {} retries attempted".format(type(err), func, retries_count, retries)
                    if retries_count >= retries:
                        if verbose:
                            _logger.exception(message)
                        raise err
                    elif verbose:
                        _logger.warning(message)

                    if cooldown:
                        _logger.debug('Cooling down for %s seconds', cooldown)
                        await asyncio.sleep(cooldown)
                    retries_count += 1
                else:
                    return result
        return inner
    return wrap


def async_retry_context_manager(*exceptions: type[Exception], retries = 3, cooldown: Optional[int] = 1,
                                verbose = True) -> Callable[[Callable[..., AsyncContextManager[RT]]], Callable[..., AsyncContextManager[RT]]]:
    def wrap(ctxman: Callable[..., AsyncContextManager[RT]]) -> type[AsyncContextManager[RT]]:
        class AsyncContextManagerWrapper(AsyncContextManager[RT]):
            def __init__(self, *args, **kwargs) -> None:
                self.__args = args
                self.__kwargs = kwargs

            @async_retry(*exceptions, retries=retries, cooldown=cooldown, verbose=verbose)
            async def __aenter__(self) -> RT:
                self.__ctxman = ctxman(*self.__args, **self.__kwargs)
                return await self.__ctxman.__aenter__()

            async def __aexit__(self, exc_type, exc, tb) -> Any:
                return await self.__ctxman.__aexit__(exc_type, exc, tb)
        return AsyncContextManagerWrapper
    return wrap


def retry(*exceptions: Type[Exception], retries: int = 3, cooldown: Optional[int] = 1,
          verbose: bool = True) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    """
    Decorate a non-async function to execute it a few times before giving up.

    :param exceptions: One or more exceptions expected during function execution, as positional arguments.
    :param retries: Number of retries of function execution.
    :param cooldown: Seconds to wait before retry. If None, does not cool down.
    :param verbose: Specifies if we should log about not successful attempts.
    :return a callable that accepts a callable, wraps it with the retry behavior, and returns a callable with the same
    signature.
    :raises: if all retries failed, or an exception other than those in the exceptions parameter was raised, the
    exception from the last try.
    """

    def wrap(func):
        @wraps(func)
        def inner(*args, **kwargs):
            _logger = logging.getLogger(__name__)
            retries_count = 0

            while True:
                try:
                    _logger.debug('About to try %s', func)
                    result = func(*args, **kwargs)
                except exceptions as err:
                    message = "Exception {} during {} execution. " \
                              "{} of {} retries attempted".format(type(err), func, retries_count, retries)

                    if retries_count >= retries:
                        verbose and _logger.exception(message)
                        raise err
                    else:
                        verbose and _logger.warning(message)

                    if cooldown:
                        _logger.debug('Cooling down for %s seconds', cooldown)
                        time.sleep(cooldown)
                    retries_count += 1
                else:
                    return result
        return inner
    return wrap


@contextmanager
def modified_environ(*remove, **update):
    """
    Temporarily updates the ``os.environ`` dictionary in-place.

    The ``os.environ`` dictionary is updated in-place so that the modification
    is sure to work in all situations. Code obtained from
    https://stackoverflow.com/questions/2059482/python-temporarily-modify-the-current-processs-environment.

    :param remove: Environment variables to remove.
    :param update: Dictionary of environment variables and values to add/update.
    """
    env = os.environ
    update = update or {}
    remove = remove or []

    # List of environment variables being updated or removed.
    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
    # Environment variables and values to restore on exit.
    update_after = {k: env[k] for k in stomped}
    # Environment variables and values to remove on exit.
    remove_after = frozenset(k for k in update if k not in env)

    try:
        env.update(update)
        [env.pop(k, None) for k in remove]
        yield
    finally:
        env.update(update_after)
        [env.pop(k) for k in remove_after]


@contextmanager
def request_get_with_retry(url: str, retries=10, cooldown: int | None = 5, timeout: float | None = 60) -> Generator[requests.Response, None, None]:
    """
    Context manager that opens a URL, retrying if unsuccessful, and generates a HTTPResponse.

    :param url: the URL to open (required).
    :param retries: the number of retries.
    :param cooldown: the cooldown period (None for no cooldown).
    :param timeout: the connection timeout (None for no timeout).
    :raises TimeoutError: if the final retry timed out.
    :raises URLError: if the final retry errored out.
    """
    logger = logging.getLogger(__name__)
    logger.debug('Connecting to URL %s', url)

    if not retries:
        retries = 0

    @retry(requests.exceptions.RequestException, retries=retries, cooldown=cooldown)
    def do_open() -> requests.Response:
        try:
            with requests.get(url, timeout=timeout) as response:
                return response
        except Exception as e:
            logger.exception(f'Error getting URL {url}')
            raise e
    response = do_open()
    try:
        yield response
    finally:
        response.close()


class DuplicateError(ValueError):
    """
    The exception raised by the check_duplicates function.
    """
    def __init__(self, msg: str, duplicate_item: Any):
        """
        Creates the error.

        :param msg: the message to display when the exception is raised.
        :param duplicate_item: the item triggering the exception.
        """
        super().__init__(msg)
        self.__duplicate_item = duplicate_item

    @property
    def duplicate_item(self) -> Any:
        """The item triggering the exception."""
        return self.__duplicate_item


def check_duplicates(itr: Iterable[Any]):
    """
    Checks an iterable for duplicate items and raises an exception if one is found.

    @param itr: an iterable.
    @raise DuplicateError if a duplicate is found.
    """
    s = set()
    for item in itr:
        if item in s:
            raise DuplicateError(f'Duplicate {item}', item)
        else:
            s.add(item)


@dataclass(frozen=True)
class TypeEnforcingFrozenDataclass:
    def __post_init__(self):
        for (name, field_type) in self.__annotations__.items():
            if not isinstance(self.__dict__[name], field_type):
                current_type = type(self.__dict__[name])
                raise TypeError(f"The field `{name}` was assigned by `{current_type}` instead of `{field_type}`")

def issubclass_silent(cls, class_or_tuple, /):
    """
    Works like the builtin issubclass, except it returns False if cls is not a
    type instead of raising a TypeError.
    """
    return isinstance(cls, type) and cls is not None and issubclass(cls, class_or_tuple)

T = TypeVar('T')

def to_type(cls_or_instance: T | Type[T] | None) -> Type[T] | None:
    if cls_or_instance is None:
        return None
    elif isinstance(cls_or_instance, type):
        return cls_or_instance
    else:
        return type(cls_or_instance)

async def queued_processing(itr: AsyncIterable[T],
                            item_processor: Callable[[T], Awaitable[None]],
                            exceptions_to_ignore: Callable[[Exception], bool] | Iterable[type[Exception]] | None = None,
                            queue_max_size=100, num_workers=10):
    """
    Places items from an async iterable into an asyncio queue for processing by the provided async item_processor.
    Specified exceptions may be suppressed. Processing order is not guaranteed unless num_workers is 1.

    :param itr: an async iterable (required).
    :param item_processor: the async item processor.
    :param exceptions_to_ignore: a function that takes an exception parameter and returns whether the exception should
    be suppressed when raised by the item processor; alternatively, an iterable of exceptions to suppress. Queue
    processing still stops, just the exceptions are not raised.
    :param queue_max_size the maximum size of the queue (default is 100).
    :param num_workers: the number of workers that take items off the queue (default is 10).
    """
    logger = logging.getLogger(__name__)
    queue = asyncio.Queue[T | None](queue_max_size)
    done_event = asyncio.Event()
    async def worker(name: str, queue: asyncio.Queue[T | None]):
        logger.debug('Starting worker %s', name)
        try:
            while True:
                try:
                    coro = await queue.get()
                    if coro is None:
                        logger.debug('Worker %s received None, exiting', name)
                        break
                    logger.debug('Worker %s processing item %s', name, coro)
                    await item_processor(coro)
                    logger.debug('Worker %s processed item %s', name, coro)
                finally:
                    queue.task_done()
        except Exception as e:
            done_event.set()
            logger.exception('Exception %s in worker %s, draining queue', e, name)
            while True:
                coro = await queue.get()
                queue.task_done()
                if coro is None:
                    break
            logger.debug('Queue drained in worker %s', name)
            raise
        finally:
            logger.debug('Exiting worker %s', name)


    logger.debug('Creating workers')
    workers: list[asyncio.Task[None]] = []
    for i in range(num_workers):
        workers.append(asyncio.create_task(worker(f'worker-{i}', queue)))
    logger.debug('Done creating workers %s', workers)
    try:
        async for coro in itr:
            if done_event.is_set():
                break
            logger.debug('Adding %s to the queue', coro)
            await queue.put(coro)
            logger.debug('Done adding %s to the queue', coro)
    finally:
        logger.debug('Cancelling workers')
        for _ in workers:
            await queue.put(None)
        await queue.join()
        exceptions_to_raise = []
        logger.debug('Gathering workers %s', workers)
        if callable(exceptions_to_ignore):
            def should_raise(r: Exception) -> bool:
                return not exceptions_to_ignore(r)
        else:
            def should_raise(r: Exception) -> bool:
                if exceptions_to_ignore:
                    return not isinstance(r, tuple(exceptions_to_ignore))
                else:
                    return True
        for r in await asyncio.gather(*workers, return_exceptions=True):
            logger.debug('Worker result %s', r)
            if isinstance(r, Exception) and should_raise(r):
                exceptions_to_raise.append(r)
        logger.debug('Exceptions to raise: %s', exceptions_to_raise)
        raise_multiple(exceptions_to_raise)


def public_methods(cls: Type) -> list[str]:
    """
    Returns a list of the provided class' public methods, defined as any callable attribute of the class that is not a
    class and whose name does not begin with an underscore.

    :param cls: the class (required).
    :return: a list of method names.
    """
    l: list[str] = []
    for method in dir(cls):
        attr = getattr(cls, method)
        if not method.startswith('_') and not inspect.isclass(attr):
            l.append(method)
    return l


async def yield_control():
    """
    Yields control from the current async task to another task.
    """
    await asyncio.sleep(0)


def raise_multiple(errors: Iterable[BaseException] | None):
    """
    Raises multiple exceptions at once, nesting them. The first exception can be caught in a try-except clause around a
    call to this function. Only the first 1000 exceptions are handled.

    FIXME: do this with ExceptionGroup after we drop support for Python 3.10, after which we can drop the 1000 limit.

    :param errors: the exceptions to raise. If the iterable has no values or is None, this function does nothing.
    """
    if not errors:
        return
    errors_ = list(errors)
    if len(errors_) > 1000:
        errors_ = errors_[:1000]
    try:
        raise errors_.pop()
    finally:
        raise_multiple(errors_)


K = TypeVar('K', bound=Any)

class AsyncioLockCache(Generic[K]):
    """
    A cache for asyncio.Lock objects with a fixed capacity, using an LRU eviction policy. K must be an immutable type.

    Attributes:
        capacity (int): The maximum number of locks the cache can hold.
        cache (OrderedDict[str, asyncio.Lock]): The internal cache storing the locks.
        ref_count (dict[str, int]): A dictionary tracking the reference count of each lock.
    """

    def __init__(self, capacity = 128) -> None:
        """
        Initialize the AsyncioLockCache with a given capacity.

        :param capacity: The maximum number of locks the cache can hold. Defaults to 128.
        """
        self.__capacity = capacity if capacity is not None else 128
        self.__cache: OrderedDict[K, asyncio.Lock] = OrderedDict()
        self.__ref_count: dict[K, int] = {}
        self.__lock = asyncio.Lock()
        self.__eviction_event = asyncio.Event()

    @property
    def capacity(self) -> int:
        """
        The capacity of the cache (read-only).
        """
        return self.__capacity

    @property
    def cache(self) -> OrderedDict[K, asyncio.Lock]:
        """
        A copy of the internal cache (read-only).
        """
        return self.__cache.copy()

    @property
    def ref_count(self) -> dict[K, int]:
        """
        A copy of the reference count dictionary (read-only).
        """
        return self.__ref_count.copy()

    async def get(self, key: K) -> asyncio.Lock:
        """
        Get the lock associated with the given key, creating a new one if necessary.

        :param key: The key for which to get the lock.
        :return: The lock associated with the given key.
        """
        async with self.__lock:
            if key in self.__cache:
                self.__cache.move_to_end(key)
                self.__ref_count[key] += 1
            else:
                if len(self.__cache) == self.__capacity:
                    await self.__evict()
                self.__cache[key] = asyncio.Lock()
                self.__ref_count[key] = 1

            return self.__cache[key]

    async def release(self, key: K) -> None:
        """
        Release the lock associated with the given key, decrementing its reference count.

        :param key: The key for which to release the lock.
        """
        async with self.__lock:
            if key in self.__ref_count:
                self.__ref_count[key] -= 1
                if self.__ref_count[key] == 0:
                    del self.__ref_count[key]
                    del self.__cache[key]
                    self.__eviction_event.set()

    async def __evict(self) -> None:
        """
        Evict the least recently used lock that is not currently in use. Will
        not return until a lock has been evicted.
        """
        while True:
            for key in list(self.__cache.keys()):
                if self.__ref_count[key] == 0:
                    del self.__ref_count[key]
                    del self.__cache[key]
                    return
            self.__eviction_event.clear()
            await self.__eviction_event.wait()


class LockManager(Generic[K]):
    """
    A context manager for acquiring and releasing asyncio locks for specific immutable objects. K must be an immutable
    type.
    """

    def __init__(self, capacity = 128) -> None:
        """
        Initialize the LockManager.

        :param capacity:  The maximum number of locks the cache can hold. Defaults to 128.
        """
        self.__lock_cache: AsyncioLockCache[K] = AsyncioLockCache(capacity=capacity)

    @asynccontextmanager
    async def lock(self, key: K):
        """
        An async context manager for acquiring and releasing a lock for a specific key.

        :param key: The key for which to acquire the lock.
        """
        lock = await self.__lock_cache.get(key)
        try:
            async with lock:
                yield
        finally:
            await self.__lock_cache.release(key)


KEY = TypeVar('KEY')
VALUE = TypeVar('VALUE')

def filter_mapping(mapping: Mapping[KEY, VALUE], keys: Collection[Any]) -> Iterator[tuple[KEY, VALUE]]:
    """
    Filter a mapping to include only the specified keys.

    :param mapping: The mapping to filter.
    :param keys: The keys to include in the filtered mapping.
    :return: The filtered mapping as an iterator of key-value pairs.
    """
    return filter(lambda item: item[0] in keys, mapping.items())


def to_http_date(dt: datetime) -> str:
    """
    Convert a datetime object to an HTTP date string.

    :param dt: The datetime object to convert.
    :return: The HTTP date string.
    """
    return formatdate(seconds_since_epoch(dt), localtime=False, usegmt=True)


async def do_nothing(*args, **kwargs) -> None:
    """
    An async no-op function that does nothing.

    :return: None.
    """
    pass



from unittest import IsolatedAsyncioTestCase, TestCase

from aiohttp.test_utils import make_mocked_request
from heaobject.aws import AWSDesktopObject, S3StorageClass, S3ArchiveDetailState
from heaobject.data import AWSS3FileObject
from heaobject.root import Permission
from heaserver.service.db import awsservicelib
from botocore.exceptions import ClientError

import heaserver.service.db.aws
from heaserver.service.db.aws import S3ObjectPermissionContext
from heaserver.service.testcase.mockaws import MockS3
from collections.abc import Sequence
from types import MethodType

class TestAWSServiceLib(TestCase):

    def test_handle_client_error_404(self):
        c = ClientError(error_response={'Error': {'Code': heaserver.service.db.aws.CLIENT_ERROR_404}}, operation_name='foo')
        self.assertEqual(404, awsservicelib.handle_client_error(c).status)

    def test_handle_client_error_no_such_bucket(self):
        c = ClientError(error_response={'Error': {'Code': heaserver.service.db.aws.CLIENT_ERROR_NO_SUCH_BUCKET}}, operation_name='foo')
        self.assertEqual(404, awsservicelib.handle_client_error(c).status)

    def test_handle_client_error_unknown(self):
        c = ClientError(error_response={'Error': {'Code': "It's wrecked"}},
                        operation_name='foo')
        self.assertEqual(500, awsservicelib.handle_client_error(c).status)


class TestArchiveStatus(IsolatedAsyncioTestCase):

    def setUp(self):
        self.request = make_mocked_request('GET', '/')
        self.request.app['HEA_db'] = MockS3(None)
        self.perm_context: S3ObjectPermissionContext = S3ObjectPermissionContext(self.request)

        async def mock_is_super_admin(self) -> bool:
            return False
        async def mock_simulate_perms(self, obj: AWSDesktopObject, actions: Sequence[str]) -> list[Permission]:
            return [Permission.VIEWER, Permission.EDITOR, Permission.DELETER]
        async def mock_is_account_owner(self) -> bool:
            return False
        self.perm_context.is_super_admin = MethodType(mock_is_super_admin, self.perm_context)
        self.perm_context._simulate_perms = MethodType(mock_simulate_perms, self.perm_context)
        self.perm_context.is_account_owner = MethodType(mock_is_account_owner, self.perm_context)

    async def test_handle_display_name_attribute_permissions_glacier_deep_archive(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.storage_class = S3StorageClass.DEEP_ARCHIVE
        obj.archive_detail_state = S3ArchiveDetailState.ARCHIVED
        obj.key = 'test_key'
        self.assertEqual([Permission.VIEWER], await self.perm_context.get_attribute_permissions(obj, 'display_name'))

    async def test_handle_display_name_attribute_permissions_glacier_instant_retrieval(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.storage_class = S3StorageClass.GLACIER_IR
        obj.archive_detail_state = S3ArchiveDetailState.NOT_ARCHIVED
        obj.key = 'test_key'
        self.assertEqual(set([Permission.VIEWER, Permission.EDITOR]), set(await self.perm_context.get_attribute_permissions(obj, 'display_name')))

    async def test_handle_display_name_attribute_permissions_standard(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.storage_class = S3StorageClass.STANDARD
        obj.key = 'test_key'
        self.assertEqual(set([Permission.VIEWER, Permission.EDITOR]), set(await self.perm_context.get_attribute_permissions(obj, 'display_name')))

    async def test_handle_display_name_attribute_permissions_unknown(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.key = 'test_key'
        self.assertEqual(set([Permission.VIEWER, Permission.EDITOR]), set(await self.perm_context.get_attribute_permissions(obj, 'display_name')))

    async def test_handle_display_name_attribute_permissions_restoring(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.key = 'test_key'
        obj.storage_class = S3StorageClass.DEEP_ARCHIVE
        obj.archive_detail_state = S3ArchiveDetailState.RESTORING
        self.assertEqual([Permission.VIEWER], await self.perm_context.get_attribute_permissions(obj, 'display_name'))

    async def test_handle_display_name_attribute_permissions_restored(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.key = 'test_key'
        obj.storage_class = S3StorageClass.DEEP_ARCHIVE
        obj.archive_detail_state = S3ArchiveDetailState.RESTORED
        self.assertEqual(set([Permission.VIEWER, Permission.EDITOR]), set(await self.perm_context.get_attribute_permissions(obj, 'display_name')))







"""
Convenience functions for handling HEAObjects.
"""

from heaserver.service.util import LockManager
from . import client
from .representor import factory as representor_factory
from .representor.error import ParseException
from heaobject.root import HEAObject, DesktopObjectTypeVar, DesktopObjectDict, DesktopObject, \
    desktop_object_type_for_name, Permission, DefaultPermissionGroup, PermissionContext, EnumWithAttrs
from heaobject.user import is_system_user
from heaobject.person import encode_group
from heaobject.error import DeserializeException
from heaobject.person import Person, Group
from heaserver.service.oidcclaimhdrs import SUB
from aiohttp import web
import logging
from typing import Union, Optional, Type
from collections.abc import Sequence, Mapping, AsyncIterator, Iterable
from yarl import URL
from asyncio import Lock


class RESTPermissionGroup(EnumWithAttrs):
    """
    Enum that maps heaobject's root.Permission enum values to get, post, put, and delete REST API calls. These are
    aliases to the heaobject package's heaobject.root.PermissionGroup enum.

    In order to execute a GET request for an object, the user must have at least one of the permissions in the
    GETTER_PERMS permission group (VIEWER, COOWNER, or EDITOR) for the object. GETTER_PERMS is an alias to heaobject's
    ACCESSOR_PERMS.

    In order to execute a POST request to create an object, the user must have at least one of the permissions in the
    POSTER_PERMS permission group (CREATOR, COOWNER) for the container in which the object will be created.
    POSTER_PERMS is an alias to heaobject's CREATOR_PERMS.

    In order to execute a PUT request to update an object, the user must have at least one of the permissions in the
    PUTTER_PERMS permission group (EDITOR, COOWNER) for the object. PUTTER_PERMS is an alias to heaobject's
    UPDATER_PERMS.

    In order to execute a DELETE request to delete an object, the user must have at least one of the permissions in the
    DELETER_PERMS permission group (DELETER, COOWNER) for the object. DELETER_PERMS is an alias to heaobject's
    DETETER_PERMS.
    """

    GETTER_PERMS = DefaultPermissionGroup.ACCESSOR_PERMS.perms
    PUTTER_PERMS = DefaultPermissionGroup.UPDATER_PERMS.perms
    POSTER_PERMS = DefaultPermissionGroup.CREATOR_PERMS.perms
    DELETER_PERMS = DefaultPermissionGroup.DELETER_PERMS.perms

    def __init__(self, perms: Iterable[Permission]):
        self.__perms = list(perms)

    @property
    def perms(self) -> list[Permission]:
        """
        The permissions that are part of the group.
        """
        return list(self.__perms)

    def get_perms_as_strs(self) -> list[str]:
        """
        Returns the permissions in the group as strings.
        """
        return list(perm.name for perm in self.__perms)

    async def has_any(self, obj: DesktopObject, context: PermissionContext) -> bool:
        """
        Returns whether the object has any of the permissions in this group.

        :param: obj: a desktop object (required).
        :param: context: the permission context (required).
        :return: True or False.
        """
        return await obj.has_permissions(self.__perms, context)

    def __contains__(self, item: Permission) -> bool:
        """
        Checks if the given permission is in the group.
        """
        return item in self.__perms


async def new_heaobject_from_type_name(request: web.Request, type_name: str) -> DesktopObject:
    """
    Creates a new HEA desktop object of the given type and populates its
    attributes from the body of a HTTP request. The form field names are
    assumed to correspond to attributes of the desktop object, and the
    attributes are set to the values in the form in order of appearance in the
    request. If the object's owner is None or no owner property was provided,
    the owner is set to the current user.

    :param request: the HTTP request.
    :param type_name: the type name of DesktopObject.
    :return: an instance of the given DesktopObject type. It is compared to the
    type of the HEA desktop object in the request body, and a
    DeserializeException is raised if the type of the HEA object is not an
    instance of this type. If the desktop object in the body has no type
    attribute, a DeserializeException is raised.
    :return: an instance of the given DesktopObject type.
    :raises DeserializeException: if creating a HEA object from the request
    body's contents failed.
    """
    _logger = logging.getLogger(__name__)
    obj = desktop_object_type_for_name(type_name)()
    return await populate_heaobject(request, obj)


async def new_heaobject_from_type(request: web.Request, type_: Type[DesktopObjectTypeVar]) -> DesktopObjectTypeVar:
    """
    Creates a new HEA desktop object of the given type and populates its
    attributes from the body of a HTTP request. The form field names are
    assumed to correspond to attributes of the desktop object, and the
    attributes are set to the values in the form in order of appearance in the
    request. If the object's owner is None or no owner property was provided,
    the owner is set to the current user.

    :param request: the HTTP request.
    :param type_: A DesktopObject type. It is compared to the type of the HEA
    desktop object in the request body, and a DeserializeException is raised if
    the type of the HEA object is not an instance of this type. If the desktop
    object in the body has no type attribute, a DeserializeException is raised.
    :return: an instance of the given DesktopObject type.
    :raises DeserializeException: if creating a HEA object from the request
    body's contents failed.
    """
    _logger = logging.getLogger(__name__)
    try:
        representor = representor_factory.from_content_type_header(request.headers['Content-Type'])
        _logger.debug('Using %s input parser', representor)
        result = await representor.parse(request)
        _logger.debug('Got dict %s', result)
        if 'type' not in result:
            raise KeyError("'type' not specified in request body. All types must be specified explicitly in the "
                           "body of the request.")
        actual_type = desktop_object_type_for_name(result['type'])
        if not issubclass(actual_type, type_):
            raise TypeError(f'Type of object in request body must be type {type_} but was {actual_type}')
        if result.get('owner', None) is None:
            result['owner'] = request.headers.get(SUB, None)
        obj = actual_type()
        obj.from_dict(result)
        return obj
    except ParseException as e:
        raise DeserializeException(str(e)) from e
    except (ValueError, TypeError) as e:
        raise DeserializeException(str(e)) from e
    except KeyError as e:
        raise DeserializeException(str(e)) from e
    except Exception as e:
        raise DeserializeException(str(e)) from e


async def populate_heaobject(request: web.Request, obj: DesktopObjectTypeVar) -> DesktopObjectTypeVar:
    """
    Populate an HEA desktop object from a POST or PUT HTTP request.

    :param request: the HTTP request. Required.
    :param obj: the HEAObject instance. Required.
    :return: the populated object.
    :raises DeserializeException: if creating a HEA object from the request body's contents failed.
    """
    _logger = logging.getLogger(__name__)
    try:
        representor = representor_factory.from_content_type_header(request.headers['Content-Type'])
        _logger.debug('Using %s input parser', representor)
        result = await representor.parse(request)
        _logger.debug('Got dict %s', result)
        obj.from_dict(result)
        return obj
    except (ParseException, ValueError) as e:
        _logger.exception('Failed to parse %s%s', obj, e)
        raise DeserializeException from e
    except Exception as e:
        _logger.exception('Got exception %s', e)
        raise DeserializeException from e


async def type_to_resource_url(request: web.Request, type_or_type_name: Union[str, Type[DesktopObject]],
                               parameters: Optional[Mapping[str, Union[Sequence[Union[int, float, complex, str]], Mapping[str, Union[int, float, complex, str]], tuple[str, Union[int, float, complex, str]], int, float, complex, str]]] = None,
                               **kwargs: Union[Sequence[Union[int, float, complex, str]], Mapping[str, Union[int, float, complex, str]], tuple[str, Union[int, float, complex, str]], int, float, complex, str]) -> str:
    """
    Use the HEA registry service to get the resource URL for accessing HEA objects of the given type. The result is
    cached to avoid unnecessary service calls.

    :param request: the HTTP request. Required.
    :param type_or_type_name: the type or type name of HEA desktop object. Required.
    :param parameters: for resource URLs with a URI template, the template parameters.
    :return: the URL string.
    :raises ValueError: if no resource URL was found in the registry or there was a problem accesing the registry
    service.
    """
    result = await client.get_resource_url(request.app, type_or_type_name, client_session=None, parameters=parameters, **kwargs)
    if result is None:
        raise ValueError(f'No resource in the registry for type {type_or_type_name}')
    return result


async def get_dict(request: web.Request, id_: str, type_or_type_name: Union[str, Type[DesktopObject]],
                   headers: Optional[Mapping[str, str]] = None) -> Optional[DesktopObjectDict]:
    """
    Gets the HEA desktop object dict with the provided id from the service for the given type, file system type,
    and file system name.

    :param request: the aiohttp request (required).
    :param id_: the id of the HEA desktop object of interest.
    :param type_or_type_name: the desktop object type or type name.
    :param headers: optional HTTP headers to use.
    :return: the requested HEA desktop object dict, or None if not found.
    :raises ValueError: if no appropriate service was found or there was a problem accessing the registry service.
    """
    url = await type_to_resource_url(request, type_or_type_name=type_or_type_name)
    return await client.get_dict(request.app, URL(url) / id_, headers)


async def get(request: web.Request, id_: str, type_: Type[DesktopObjectTypeVar],
              headers: Optional[Mapping[str, str]] = None) -> Optional[DesktopObjectTypeVar]:
    """
    Gets the HEA desktop object with the provided id from the service for the given type, file system type, and file
    system name.

    :param request: the aiohttp request (required).
    :param id_: the id of the HEA desktop object of interest.
    :param type_: the desktop object type.
    :param headers: optional HTTP headers to use.
    :return: the requested HEA desktop object, or None if not found.
    :raises ValueError: if no appropriate service was found or there was a problem accessing the registry service.
    """
    url = await type_to_resource_url(request, type_or_type_name=type_)

    return await client.get(request.app, URL(url) / id_, type_, headers)


async def get_all(request: web.Request, type_: Type[DesktopObjectTypeVar],
                  headers: Optional[Mapping[str, str]] = None) -> AsyncIterator[DesktopObjectTypeVar]:
    """
    Async iterator for all HEA desktop objects from the service for the given type, file system type, and file system
    name.

    :param request: the aiohttp request (required).
    :param type_: the desktop object type.
    :param headers: optional HTTP headers to use.
    :return: an async iterator with the requested desktop objects.
    :raises ValueError: if no appropriate service was found or there was a problem accessing the registry service.
    """
    url = await type_to_resource_url(request, type_or_type_name=type_)

    return client.get_all(request.app, url, type_, headers)


def desktop_object_type_or_type_name_to_type(type_or_type_name: str | type[DesktopObject], default_type: type[DesktopObject] | None = None) -> type[DesktopObject]:
    """
    Takes a variable that may contain either a DesktopObject type or type name, and return a type. If a type is passed
    in, and it is a subclass of DesktopObject, it will be returned as-is. If a type name is passed in, its corresponding
    type will be returned if the type is a subclass of DesktopObject.

    :param type_or_type_name: the type or type name. Required.
    :param default_type: what to return if type_or_type_name is None. If omitted, None is returned.
    """
    if default_type is not None and not issubclass(default_type, DesktopObject):
        raise TypeError('default_type is defined and not a DesktopObject')
    if isinstance(type_or_type_name, type):
        if not issubclass(type_or_type_name, DesktopObject):
            raise TypeError(f'type_or_type_name not a DesktopObject')
        if issubclass(type_or_type_name, DesktopObject):
            result_ = type_or_type_name
        else:
            raise TypeError('type_or_type_name not a DesktopObject')
    else:
        file_system_type_ = desktop_object_type_for_name(type_or_type_name)
        if not issubclass(file_system_type_, DesktopObject):
            raise TypeError(f'file_system_type_or_type_name is a {file_system_type_} not a DesktopObject')
        else:
            result_ = file_system_type_
    if result_ is None:
        return default_type
    else:
        return result_


def type_or_type_name_to_type(type_or_type_name: str | type[HEAObject], default_type: type[HEAObject] | None = None) -> type[HEAObject]:
    """
    Takes a variable that may contain either a HEAObject type or type name, and return a type. If a type is passed
    in, and it is a subclass of HEAObject, it will be returned as-is. If a type name is passed in, its corresponding
    type will be returned if the type is a subclass of HEAObject.

    :param type_or_type_name: the type or type name. Required.
    :param default_type: what to return if type_or_type_name is None. If omitted, None is returned.
    """
    if default_type is not None and not issubclass(default_type, HEAObject):
        raise TypeError('default_type is defined and not a HEAObject')
    if isinstance(type_or_type_name, type):
        if not issubclass(type_or_type_name, HEAObject):
            raise TypeError(f'type_or_type_name not a HEAObject')
        if issubclass(type_or_type_name, HEAObject):
            result_ = type_or_type_name
        else:
            raise TypeError('type_or_type_name not a HEAObject')
    else:
        file_system_type_ = desktop_object_type_for_name(type_or_type_name)
        if not issubclass(file_system_type_, HEAObject):
            raise TypeError(f'file_system_type_or_type_name is a {file_system_type_} not a HEAObject')
        else:
            result_ = file_system_type_
    if result_ is None:
        return default_type
    else:
        return result_


def display_name_for_multiple_objects(objs: Sequence[DesktopObject]) -> str:
    """
    Returns a user-friendly display name for multiple desktop objects. For one
    object, it will return just the object's display name.

    :param objs: a sequence of desktop objects
    :raises ValueError: if objs is None or an empty sequence.
    """
    if objs is None:
        raise ValueError('objs cannot be None')
    match len(objs):
        case 0:
            raise ValueError('No objects in objs')
        case 1:
            return objs[0].display_name
        case 2:
            return f'{objs[0].display_name} and 1 other object'
        case _:
            return f'{objs[0].display_name} and {len(objs) - 1} other objects'


class HEAServerPermissionContext(PermissionContext):
    """
    A context for checking permissions on HEA desktop objects that uses the heaserver-people service to access the
    current user's groups and checks group-level permissions in addition to permissions that are assigned directly to
    the user.
    """

    def __init__(self, sub: str, request: web.Request, **kwargs):
        super().__init__(sub, **kwargs)
        if request is None:
            raise ValueError('request cannot be None')
        self.__request = request
        self.__groups: list[str] | None = None
        self.__group_to_group_id: dict[str, str | None] = {}
        self.__group_to_group_id_lock = LockManager[str]()
        self.__groups_lock = Lock()

    @property
    def request(self) -> web.Request:
        """The HTTP request."""
        return self.__request

    async def get_groups(self) -> list[str]:
        """
        Gets the group ids for the current user. This implementation consults the REST APIs of the service responsible
        for Persons.

        :return: the group ids for the current user.
        """
        async with self.__groups_lock:
            if self.__groups is None:
                sub = self.sub
                if is_system_user(sub):
                    self.__groups = []
                else:
                    url = await type_to_resource_url(self.request, Person)
                    groups = await client.get_all_list(self.request.app, URL(url) / sub / 'groups', Group,
                                                    headers={SUB: sub})
                    assert all(group.id is not None for group in groups), f'One or more groups in {groups} has no id'
                    self.__groups = [group.id for group in groups if group.id is not None]
            return self.__groups

    async def group_id_from(self, group: str) -> str:
        """
        Gets the group id for a group. This implementation consults the REST APIs of the service responsible for
        Groups.

        :param group: the group (required).
        :return: the group id.
        :raises ValueError: if the group was not found.
        """
        async with self.__group_to_group_id_lock.lock(group):
            if group in self.__group_to_group_id:
                group_id = self.__group_to_group_id[group]
                if group_id is not None:
                    return group_id
                group_: Group | None = None
            else:
                url = await type_to_resource_url(self.request, Group)
                group_ = await client.get(self.request.app, URL(url) / 'byname' / encode_group(group), Group,
                                          headers={SUB: self.sub})
                self.__group_to_group_id[group] = group_.id if group_ is not None else None
            if group_ is None:
                raise ValueError(f'Group {group} not found')
            assert group_.id is not None, f'Group {group_} has no id'
            return group_.id

    async def can_create(self, desktop_object_type: type[DesktopObject]) -> bool:
        """
        Checks if the current user can create a desktop object of the given type. This implementation consults the
        REST APIs of the service responsible for registry Components.

        :param desktop_object_type: the type of the desktop object (required).
        :return: True if the user can create the object, False otherwise.
        """
        component = await client.get_component(self.request.app, desktop_object_type)
        if component is None or (resource := component.get_resource(desktop_object_type.get_type_name())) is None:
            raise ValueError(f'Invalid desktop object type {desktop_object_type}')
        return resource.manages_creators and await resource.is_creator(self)
